# Copyright (c) 2024, Manfred Moitzi
# License: MIT License
import pytest

import ezdxf
import ezdxf.xref
from ezdxf.document import Drawing
from ezdxf.entities import Solid3d
from ezdxf.render import forms


def source_doc_R2007() -> Drawing:
    doc = ezdxf.new("R2007")
    msp = doc.modelspace()
    forms.cube().render_3dsolid(msp)
    return doc


def source_doc_R2013() -> Drawing:
    doc = ezdxf.new("R2013")
    msp = doc.modelspace()
    forms.cube().render_3dsolid(msp)
    return doc


def test_load_acis_from_R2007():
    target_doc = ezdxf.new("R2007")
    source_doc = source_doc_R2007()
    source_msp = source_doc.modelspace()
    source_cube = source_msp[0]
    assert isinstance(source_cube, Solid3d)

    ezdxf.xref.load_modelspace(source_doc, target_doc)

    target_msp = target_doc.modelspace()
    loaded_cube = target_msp[0]
    assert isinstance(loaded_cube, Solid3d)
    assert loaded_cube.sat == source_cube.sat


def test_load_acis_from_R2013():
    target_doc = ezdxf.new("R2013")
    source_doc = source_doc_R2013()
    source_msp = source_doc.modelspace()
    source_cube = source_msp[0]
    assert isinstance(source_cube, Solid3d)

    ezdxf.xref.load_modelspace(source_doc, target_doc)

    target_msp = target_doc.modelspace()
    loaded_cube = target_msp[0]
    assert isinstance(loaded_cube, Solid3d)
    sab = loaded_cube.sab
    assert len(sab) > 1
    assert sab == source_cube.sab


def test_load_acis_from_2007_into_2013():
    target_doc = ezdxf.new("R2013")  # SAB

    source_doc = source_doc_R2007()
    ezdxf.xref.load_modelspace(source_doc, target_doc)

    target_msp = target_doc.modelspace()
    loaded_cube = target_msp[0]
    assert isinstance(loaded_cube, Solid3d)
    assert len(loaded_cube.sat) == 0, "SAT data removed"
    assert len(loaded_cube.sab) == 0


def test_load_acis_from_2013_into_2007():
    target_doc = ezdxf.new("R2007")  # SAT

    source_doc = source_doc_R2013()
    ezdxf.xref.load_modelspace(source_doc, target_doc)

    target_msp = target_doc.modelspace()
    loaded_cube = target_msp[0]
    assert isinstance(loaded_cube, Solid3d)
    assert len(loaded_cube.sat) == 0
    assert len(loaded_cube.sab) == 0, "SAB data removed"


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2020, Manfred Moitzi
#  License: MIT License
# Test only basic features of Cython implementation,
# Full testing and compatibility check with Python implementation
# is located in test suite 630a.

import pytest
import math

bezier = pytest.importorskip("ezdxf.acc.bezier4p")
Bezier4P = bezier.Bezier4P
from ezdxf.acc.vector import Vec3, Vec2
from ezdxf.acc.matrix44 import Matrix44

# check functions:
from ezdxf.math._bezier4p import (
    cubic_bezier_arc_parameters,
    cubic_bezier_from_arc,
    cubic_bezier_from_ellipse,
)
from ezdxf.math.ellipse import ConstructionEllipse

POINTS = Vec2.list([(0, 0), (1, 0), (1, 1), (0, 1)])


def test_default_constructor():
    c = Bezier4P(POINTS)
    assert c.control_points == Vec3.tuple(POINTS)


@pytest.fixture
def curve():
    return Bezier4P(POINTS)


def test_point(curve):
    assert curve.point(0.5) == (0.75, 0.5)


def test_tangent(curve):
    assert curve.tangent(0.5) == (0.0, 1.5)


def test_approximate(curve):
    points = curve.approximate(10)
    assert len(points) == 11


def test_flattening(curve):
    points = curve.flattening(0.01)
    assert len(points) == 17


def test_approximated_length(curve):
    assert curve.approximated_length(32) == 1.999232961352048


def test_reverse(curve):
    r = curve.reverse()
    assert r.control_points == Vec3.tuple(reversed(POINTS))


def test_transform(curve):
    m = Matrix44.translate(1, 1, 0)
    r = curve.transform(m)
    assert r.control_points == Vec3.tuple([(1, 1), (2, 1), (2, 2), (1, 2)])


@pytest.mark.parametrize(
    "s, e",
    [
        (0, math.pi),
        (math.pi, math.tau),
        (0, math.tau),
    ],
)
def test_correctness_arc_parameters(s, e):
    expected = list(cubic_bezier_arc_parameters(s, e))
    result = list(bezier.cubic_bezier_arc_parameters(s, e))
    assert result == expected


@pytest.mark.parametrize(
    "s, e",
    [
        (0, 180),
        (180, 360),
        (0, 360),
    ],
)
def test_correctness_bezier_from_arc(s, e):
    expected = list(
        cubic_bezier_from_arc(
            center=Vec3(1, 2),
            start_angle=s,
            end_angle=e,
        )
    )
    result = list(
        bezier.cubic_bezier_from_arc(
            center=Vec3(1, 2),
            start_angle=s,
            end_angle=e,
        )
    )
    for e, r in zip(expected, result):
        assert e.control_points == r.control_points


@pytest.mark.parametrize(
    "s, e",
    [
        (0, math.pi),
        (math.pi, math.tau),
        (0, math.tau),
    ],
)
def test_correctness_bezier_from_ellipse(s, e):
    ellipse = ConstructionEllipse(
        center=(1, 2),
        major_axis=(2, 0),
        ratio=0.5,
        start_param=s,
        end_param=e,
    )
    expected = list(cubic_bezier_from_ellipse(ellipse))
    result = list(bezier.cubic_bezier_from_ellipse(ellipse))
    for e, r in zip(expected, result):
        for ep, rp in zip(e.control_points, r.control_points):
            assert ep.isclose(rp)


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2020-2023, Manfred Moitzi
#  License: MIT License
import pytest
import pathlib
import random
from ezdxf import recover
from ezdxf.audit import AuditError
from ezdxf.lldxf.tagger import tag_compiler, ascii_tags_loader

DATA = pathlib.Path(__file__).parent / "data"
RECOVER1 = "recover01.dxf"
RECOVER2 = "recover02.dxf"
CC_DXFLIB = "cc_dxflib.dxf"
EMPTY_HANDLES = "empty_handles.dxf"


def fullpath(name):
    filepath = DATA / name
    if not filepath.exists():
        pytest.skip(f"File {filepath} not found.")
    return str(filepath)


@pytest.fixture
def tags01():
    filename = fullpath(RECOVER1)
    tool = recover.Recover()
    with open(filename, "rb") as fp:
        return list(tool.load_tags(fp, errors="ignore"))


def test_bytes_loader():
    filename = fullpath(RECOVER1)
    with open(filename, "rb") as fp:
        tags = list(recover.bytes_loader(fp))
    assert len(tags) == 14736


def test_safe_tag_loader():
    filename = fullpath(RECOVER1)
    with open(filename, "rt", encoding="cp1252") as fp:
        expected = list(tag_compiler(iter(ascii_tags_loader(fp))))

    with open(filename, "rb") as fp:
        tags = list(recover.safe_tag_loader(fp))

    assert len(tags) == len(expected)
    assert tags[:100] == tags[:100]


def test_rebuild_sections(tags01):
    tool = recover.Recover()
    sections = tool.rebuild_sections(tags01)
    expected = sum(int(tag == (0, "SECTION")) for tag in tags01)
    orphans = sections.pop()
    assert len(sections) == expected
    assert len(orphans) == 4


def test_build_section_dict(tags01):
    tool = recover.Recover()
    sections = tool.rebuild_sections(tags01)
    tool.load_section_dict(sections)
    assert len(tool.section_dict) == 2
    header = tool.section_dict["HEADER"][0]
    assert len(header) == 6
    assert header[0] == (0, "SECTION")
    assert header[1] == (2, "HEADER")
    assert len(tool.section_dict["ENTITIES"]) == 1505


def test_readfile_recover01_dxf():
    doc, auditor = recover.readfile(fullpath(RECOVER1))
    assert doc.dxfversion == "AC1009"
    assert auditor.has_errors is False


@pytest.fixture
def tags02():
    filename = fullpath(RECOVER2)
    tool = recover.Recover()
    with open(filename, "rb") as fp:
        return list(tool.load_tags(fp, errors="ignore"))


def test_rebuild_tables(tags02):
    recover_tool = recover.Recover()
    sections = recover_tool.rebuild_sections(tags02)
    recover_tool.load_section_dict(sections)
    tables = recover_tool.section_dict.get("TABLES")
    random.shuffle(tables)

    tables = recover_tool.rebuild_tables(tables)
    assert tables[0] == [(0, "SECTION"), (2, "TABLES")]
    assert tables[1][0] == (0, "TABLE")
    assert tables[1][1] == (2, "VPORT")
    assert tables[2][0] == (0, "VPORT")
    assert tables[3][0] == (0, "ENDTAB")

    assert tables[4][0] == (0, "TABLE")
    assert tables[4][1] == (2, "LTYPE")
    assert tables[5][0] == (0, "LTYPE")
    assert tables[8][0] == (0, "ENDTAB")

    assert tables[-5][0] == (0, "TABLE")
    assert tables[-5][1] == (2, "BLOCK_RECORD")
    assert tables[-4][0] == (0, "BLOCK_RECORD")
    assert tables[-1][0] == (0, "ENDTAB")


def test_readfile_recover02_dxf():
    doc, auditor = recover.readfile(fullpath(RECOVER2))
    assert doc.dxfversion == "AC1032"
    assert auditor.has_errors is False

    # Auditor should restore deleted BLOCK-RECORD table head:
    blkrec_head = doc.block_records.head
    assert blkrec_head.dxf.handle is not None
    assert blkrec_head.dxf.handle in doc.entitydb

    # Auditor should update/fix BLOCK_RECORD entries owner handle:
    for entry in doc.block_records:
        assert (
            entry.dxf.owner == blkrec_head.dxf.handle
        ), "Auditor() should update table-entry owner handle."

    # Auditor should restore invalid VPORT table-head owner handle:
    vport_head = doc.viewports.head
    assert (
        vport_head.dxf.owner == "0"
    ), "Auditor() should repair invalid table-head owner handle."

    # Auditor should fix invalid VPORT table-entry owner handle:
    vport = doc.viewports.get("*Active")[0]
    assert (
        vport.dxf.owner == vport_head.dxf.handle
    ), "Auditor() should update table-entry owner handle."


def test_read_cc_dxflib_file():
    doc, auditor = recover.readfile(fullpath(CC_DXFLIB))
    codes = {fix.code for fix in auditor.fixes}
    assert AuditError.REMOVED_UNSUPPORTED_SECTION in codes
    assert AuditError.REMOVED_UNSUPPORTED_TABLE in codes
    msp = doc.modelspace()
    polyline = msp.query("POLYLINE").first
    assert polyline is not None


def test_readfile_empty_handles_dxf():
    doc, auditor = recover.readfile(fullpath(EMPTY_HANDLES))
    msp = doc.modelspace()
    assert doc.dxfversion == "AC1009"
    assert auditor.has_errors is False
    assert len(msp.query("LINE")) == 8
    assert len(msp.query("*[layer=='GEOMETRY-CUTTING']")) == 4
    assert len(msp.query("*[layer=='GEOMETRY-ENGRAVING']")) == 4


def test_decode_dxf_unicode_automatically():
    # The test file contains the layer name: "Tschüss mit \U+00FC"
    doc, _ = recover.readfile(fullpath("dxf_unicode.dxf"))
    assert any("Tschüss mit ü" == layer.dxf.name for layer in doc.layers) is True


def test_recover_layout_broken_links_1():
    """
    Links broken in one direction.

    LAYOUT entities have invalid handles to block records but BLOCK_RECORDs have
    valid handles to LAYOUT entities.
    """
    doc, _ = recover.readfile(fullpath("layout_broken_links.dxf"))
    assert len(doc.layouts) == 3
    msp = doc.modelspace()
    assert msp.block_record.dxf.name == "*Model_Space"
    layout1 = doc.layouts.get("Layout1")
    assert layout1.block_record.dxf.name == "*Paper_Space"
    layout2 = doc.layouts.get("Layout2")
    assert layout2.block_record.dxf.name == "*Paper_Space0"


def test_recover_layout_broken_links_2():
    """
    Links broken in both directions.
    
    LAYOUT entities have invalid handles to block records and BLOCK_RECORDs have no 
    LAYOUT handles.
    """
    doc, _ = recover.readfile(fullpath("layout_broken_links_2.dxf"))
    assert len(doc.layouts) == 3
    msp = doc.modelspace()
    assert msp.block_record.dxf.name == "*Model_Space"
    layout1 = doc.layouts.get("Layout1")
    assert layout1.block_record.dxf.name == "*Paper_Space"
    layout2 = doc.layouts.get("Layout2")
    assert layout2.block_record.dxf.name == "*Paper_Space0"


def test_recover_layout_missing_block_record():
    """BLOCK_RECORD for Layout2 does not exist."""
    doc, _ = recover.readfile(fullpath("layout_missing_block_record.dxf"))
    assert len(doc.layouts) == 3
    layout1 = doc.layouts.get("Layout1")
    assert layout1.block_record.dxf.name == "*Paper_Space"

    layout2 = doc.layouts.get("Layout2")
    block_record2 = layout2.block_record
    assert block_record2.dxf.name == "*Paper_Space0"

    block2 = doc.blocks.get(block_record2.dxf.name)
    assert block2.name == block_record2.dxf.name


def test_recover_layout_missing_block_definition():
    """BLOCK definition for Layout2 does not exist."""
    doc, _ = recover.readfile(fullpath("layout_missing_block_definition.dxf"))
    assert len(doc.layouts) == 3
    block_name = "*Paper_Space0"

    block_def = doc.blocks.get(block_name)
    block_record = block_def.block_record
    assert block_record.dxf.name == block_name

    layout2 = doc.layouts.get("Layout2")
    assert layout2.block_record_name == block_name
    assert layout2.block_record is block_record


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License

import pytest

import ezdxf
from ezdxf.entities.mtext import (
    load_columns_from_xdata,
    MText,
    ColumnType,
    MTextColumns,
)
from ezdxf.lldxf.extendedtags import ExtendedTags
from ezdxf.lldxf.tagwriter import TagCollector
from ezdxf.lldxf import const
from ezdxf.entities.xdata import XData
from ezdxf.math import Vec3

DYNAMIC_MANUAL_HEIGHT = """0
MTEXT
5
9C
330
1F
100
AcDbEntity
8
0
100
AcDbMText
1001
ACAD
1000
ACAD_MTEXT_COLUMN_INFO_BEGIN
1070
75
1070
2
1070
79
1070
0
1070
76
1070
3
1070
78
1070
0
1070
48
1040
50.0
1070
49
1040
12.5
1070
50
1070
3
1040
164.8
1040
154.3
1040
0.0
1000
ACAD_MTEXT_COLUMN_INFO_END
1000
ACAD_MTEXT_COLUMNS_BEGIN
1070
47
1070
3
1005
238
1005
239
1000
ACAD_MTEXT_COLUMNS_END
"""

DYNAMIC_AUTO_HEIGHT = """0
MTEXT
5
A2
330
1F
100
AcDbEntity
8
0
100
AcDbMText
1001
ACAD
1000
ACAD_MTEXT_COLUMN_INFO_BEGIN
1070
75
1070
2
1070
79
1070
1
1070
76
1070
3
1070
78
1070
0
1070
48
1040
50.0
1070
49
1040
12.5
1000
ACAD_MTEXT_COLUMN_INFO_END
1000
ACAD_MTEXT_COLUMNS_BEGIN
1070
47
1070
3
1005
23C
1005
23D
1000
ACAD_MTEXT_COLUMNS_END
1000
ACAD_MTEXT_DEFINED_HEIGHT_BEGIN
1070
46
1040
158.1
1000
ACAD_MTEXT_DEFINED_HEIGHT_END
"""

STATIC = """0
MTEXT
5
9D
330
1F
100
AcDbEntity
8
0
100
AcDbMText
1001
ACAD
1000
ACAD_MTEXT_COLUMN_INFO_BEGIN
1070
75
1070
1
1070
79
1070
0
1070
76
1070
3
1070
78
1070
0
1070
48
1040
50.0
1070
49
1040
12.5
1000
ACAD_MTEXT_COLUMN_INFO_END
1000
ACAD_MTEXT_COLUMNS_BEGIN
1070
47
1070
3
1005
23A
1005
23B
1000
ACAD_MTEXT_COLUMNS_END
1000
ACAD_MTEXT_DEFINED_HEIGHT_BEGIN
1070
46
1040
150.0
1000
ACAD_MTEXT_DEFINED_HEIGHT_END
"""

NO_COLUMN_INFO = """0
MTEXT
5
9D
330
1F
100
AcDbEntity
8
0
100
AcDbMText
1001
ACAD
"""


# The dynamic auto height and static types are very similar, the difference is
# important for CAD applications, but not for the DXF format itself.


def get_xdata(txt: str):
    tags = ExtendedTags.from_text(txt)
    return XData(tags.xdata)


def test_load_dynamic_cols_manual_height():
    """Every column can have a different height.

    The linked MTEXT entities do NOT have the "defined_height" stored in the
    XDATA section ACAD. The only source for the column height is the
    MTextColumns.heights list.

    The linked MTEXT entities do not have an ACAD section at all.

    """
    xdata = get_xdata(DYNAMIC_MANUAL_HEIGHT)
    cols = load_columns_from_xdata(MText().dxf, xdata)
    assert cols.count == 3
    assert cols.column_type == ColumnType.DYNAMIC
    assert cols.auto_height is False
    assert cols.reversed_column_flow is False
    assert cols.defined_height == 0.0, "not defined if auto_height is False"
    assert cols.width == 50.0
    assert cols.gutter_width == 12.5
    assert cols.total_width == 175.0  # 3 * 50 + 2 * 12.5
    # total_height = max(heights) even if the last column is the tallest
    assert cols.total_height == 164.8  # max height
    assert len(cols.linked_handles) == 2
    assert len(cols.heights) == 3  # not stored the main MTEXT entity
    assert cols.heights[-1] == 0.0, "last column height has to be 0.0"
    assert cols.heights == [164.8, 154.3, 0.0]


def test_load_dynamic_cols_with_auto_height():
    """All columns have the same column height.

    Each linked MTEXT has the "defined_height" stored in the XDATA section ACAD.

    """
    xdata = get_xdata(DYNAMIC_AUTO_HEIGHT)
    cols = load_columns_from_xdata(MText().dxf, xdata)
    # Count is a calculated value, group code 72 (column height count) is 0!
    assert cols.count == 3
    assert cols.column_type == ColumnType.DYNAMIC
    assert cols.auto_height is True
    assert cols.reversed_column_flow is False
    assert cols.defined_height == 158.1, "required if auto_height is True"
    assert cols.width == 50.0
    assert cols.gutter_width == 12.5
    assert cols.total_width == 175.0  # 3 * 50 + 2 * 12.5
    assert cols.total_height == 158.1
    assert len(cols.linked_handles) == 2
    assert len(cols.heights) == 0


def test_load_static_cols():
    """All columns have the same column height.

    Each linked MTEXT has the "defined_height" stored in the XDATA section ACAD.

    """
    xdata = get_xdata(STATIC)
    cols = load_columns_from_xdata(MText().dxf, xdata)
    assert cols.count == 3
    assert cols.column_type == ColumnType.STATIC
    assert cols.auto_height is False
    assert cols.reversed_column_flow is False
    assert cols.defined_height == 150.0, "required for static columns"
    assert cols.width == 50.0
    assert cols.gutter_width == 12.5
    assert cols.total_width == 175.0  # 3 * 50 + 2 * 12.5
    assert cols.total_height == 150.0
    assert len(cols.linked_handles) == 2
    assert len(cols.heights) == 0


def test_mtext_without_column_info():
    xdata = get_xdata(NO_COLUMN_INFO)
    cols = load_columns_from_xdata(MText().dxf, xdata)
    assert cols is None


def make_mtext(txt: str) -> MText:
    mtext = MText()
    xdata = get_xdata(txt)
    mtext._columns = load_columns_from_xdata(mtext.dxf, xdata)
    return mtext


@pytest.mark.parametrize(
    "txt",
    [STATIC, DYNAMIC_AUTO_HEIGHT, DYNAMIC_MANUAL_HEIGHT],
    ids=["STATIC", "DYNAMIC_AUTO", "DYNAMIC_MANUAL"],
)
def test_set_columns_xdata(txt):
    """Create column data as XDATA for DXF versions before R2018."""
    mtext = make_mtext(txt)
    mtext.set_column_xdata()
    acad = mtext.xdata.get("ACAD")
    expected_xdata = get_xdata(txt)
    assert acad == expected_xdata.get("ACAD")


def new_mtext_with_linked_columns(count=3) -> MText:
    columns = MTextColumns()
    columns.count = count
    columns.width = 10
    columns.gutter_width = 0.5
    columns.defined_height = 50
    mtext = MText.new()
    mtext.setup_columns(columns, linked=True)
    return mtext


def test_create_new_mtext_with_linked_columns():
    mtext = new_mtext_with_linked_columns(3)
    columns = mtext.columns
    assert columns.column_type == ColumnType.STATIC
    assert len(columns.linked_columns) == 2
    assert len(columns.heights) == 0, "all columns have the same defined height"
    assert mtext.dxf.width == columns.width
    assert columns.total_height == columns.defined_height
    assert columns.total_width == 31

    # default location and text direction:
    assert mtext.dxf.insert == (0, 0, 0)
    assert mtext.dxf.text_direction == (1, 0, 0)

    for column in columns.linked_columns:
        assert column.columns is None
        assert column.dxf.width == columns.width
        assert column.dxf.defined_height == columns.defined_height
        assert column.dxf.text_direction == (1, 0, 0)

    col0, col1 = columns.linked_columns
    assert col0.dxf.insert == (10.5, 0, 0)
    assert col1.dxf.insert == (21, 0, 0)


def test_destroy_mtext_with_linked_columns():
    mtext = new_mtext_with_linked_columns(3)
    columns = mtext.columns
    mtext.destroy()
    assert mtext.is_alive is False
    for col in columns.linked_columns:
        assert col.is_alive is False


@pytest.fixture(scope="module")
def doc():
    return ezdxf.new("R2013")


def test_add_mtext_with_linked_columns_to_entitydb(doc):
    db = doc.entitydb
    mtext = new_mtext_with_linked_columns(3)
    count = len(db)
    doc.entitydb.add(mtext)
    assert len(db) == count + 3, "should add all columns to the entitydb"
    columns = mtext.columns
    for col in columns.linked_columns:
        assert col.dxf.handle is not None
        assert col.dxf.handle in db, "should be stored in the entity db"


def test_add_mtext_with_linked_columns_to_msp(doc):
    msp = doc.modelspace()
    mtext = new_mtext_with_linked_columns(3)
    count = len(msp)
    msp.add_entity(mtext)
    assert len(msp) == count + 1, "should only add the main column to the msp"

    columns = mtext.columns
    db = doc.entitydb
    for col in columns.linked_columns:
        assert col.dxf.handle is not None
        assert col.dxf.handle in db, "should be stored in the entity db"
        assert col.dxf.owner is None, "should not be stored in an entity space"


def test_delete_mtext_with_linked_columns_from_entitydb(doc):
    """Delete MTEXT by EntityDB.delete_entity()."""
    db = doc.entitydb
    mtext = new_mtext_with_linked_columns(3)
    doc.entitydb.add(mtext)
    count = len(db)
    doc.entitydb.delete_entity(mtext)
    db.purge()
    assert len(db) == count - 3, "should delete all columns from entitydb"


def test_discard_mtext_with_linked_columns_from_entitydb(doc):
    """Discard MTEXT columns from entitydb, but do not destroy the entities."""
    db = doc.entitydb
    mtext = new_mtext_with_linked_columns(3)
    doc.entitydb.add(mtext)
    count = len(db)
    doc.entitydb.discard(mtext)
    assert len(db) == count - 3, "should discard all columns from entitydb"
    assert mtext.is_alive is True
    for col in mtext.columns.linked_columns:
        assert col.is_alive is True


def test_destroy_mtext_with_linked_columns_from_entitydb(doc):
    """Delete MTEXT from EntityDB using MText.destroy()."""
    db = doc.entitydb
    mtext = new_mtext_with_linked_columns(3)
    doc.entitydb.add(mtext)
    count = len(db)
    mtext.destroy()
    db.purge()
    assert len(db) == count - 3, "should delete all columns from entitydb"


def test_copy_mtext_with_linked_columns():
    mtext = new_mtext_with_linked_columns(3)
    mtext2 = mtext.copy()
    columns = mtext.columns
    columns2 = mtext2.columns
    assert columns is not columns2
    assert len(columns2.linked_columns) == 2

    for col1, col2 in zip(columns.linked_columns, columns2.linked_columns):
        assert col1 is not col2


def test_transform_mtext_with_linked_columns():
    mtext = new_mtext_with_linked_columns(3)
    offset = Vec3(1, 2, 3)
    mtext2 = mtext.copy()
    mtext2.translate(*offset.xyz)
    assert mtext2.dxf.insert.isclose(mtext.dxf.insert + offset)
    for col1, col2 in zip(
        mtext.columns.linked_columns, mtext2.columns.linked_columns
    ):
        assert col2.dxf.insert.isclose(col1.dxf.insert + offset)


def test_scale_mtext_with_linked_columns():
    mtext = new_mtext_with_linked_columns(3)
    mtext.scale(2, 3, 1)
    assert mtext.dxf.width == 20
    columns = mtext.columns
    assert columns.width == 20
    assert columns.gutter_width == 1
    assert columns.total_width == 62
    assert columns.defined_height == 150
    for column in columns.linked_columns:
        assert column.dxf.width == 20


def test_sync_common_attribs_of_linked_columns():
    mtext = new_mtext_with_linked_columns(3)
    mtext.dxf.style = "NewStyle"
    # manual sync required - this is done automatically at DXF export:
    mtext.sync_common_attribs_of_linked_columns()
    for column in mtext.columns.linked_columns:
        assert column.dxf.style == "NewStyle"


def test_remove_dependencies_from_mtext_and_linked_columns(doc):
    mtext = new_mtext_with_linked_columns(3)
    mtext.dxf.style = "StyleNotExist"
    mtext.sync_common_attribs_of_linked_columns()
    mtext.remove_dependencies(doc)
    assert mtext.dxf.style == "Standard"
    for column in mtext.columns.linked_columns:
        assert column.dxf.handle is None
        assert column.dxf.owner is None
        assert column.dxf.style == "Standard"


class TestPreprocessDXFExport:
    collector = TagCollector(dxfversion=const.DXF2000)

    def test_successful_preprocessing(self):
        mtext = new_mtext_with_linked_columns(3)
        assert mtext.preprocess_export(self.collector) is True

    def test_accept_invalid_column_count(self):
        # Changed 2021-10-17:
        # AutoCAD accepts a different column count as linked MTEXT entities,
        # see issue #558. This occurs in real world DXF files and the behavior
        # until this fix, removed valid MTEXT entities from DXF files if they
        # were exported (saveas).
        mtext = new_mtext_with_linked_columns(3)
        mtext._columns.count = 4
        assert mtext.preprocess_export(self.collector) is True

    def test_fail_for_destroyed_columns(self):
        mtext = new_mtext_with_linked_columns(3)
        mtext.columns.linked_columns[0].destroy()
        assert mtext.preprocess_export(self.collector) is False


class TestGetColumnContent:
    def new_mtext(self):
        mtext = new_mtext_with_linked_columns(3)
        mtext.text = "Line1\\PLine2\\P"
        mtext._columns.linked_columns[0].text = "Line3\\PLine4\\P"
        mtext._columns.linked_columns[1].text = "Line5\\PLine6\\P"
        return mtext

    def test_get_raw_content_of_all_columns(self):
        mtext = self.new_mtext()
        content = mtext.all_columns_raw_content()
        assert content == "Line1\\PLine2\\PLine3\\PLine4\\PLine5\\PLine6\\P"

    def test_get_plain_text_of_all_columns_as_string(self):
        mtext = self.new_mtext()
        content = mtext.all_columns_plain_text(split=False)
        assert content == "Line1\nLine2\nLine3\nLine4\nLine5\nLine6\n"

    def test_get_plain_text_of_all_columns_as_list_of_strings(self):
        mtext = self.new_mtext()
        content = mtext.all_columns_plain_text(split=True)
        assert content == "Line1 Line2 Line3 Line4 Line5 Line6".split()


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2019-2024 Manfred Moitzi
# License: MIT License
import random

import pytest
import math
import numpy as np

from ezdxf.math import (
    Vec3,
    angle_to_param,
    ConstructionEllipse,
    ellipse_param_span,
)


def test_default_init():
    e = ConstructionEllipse()
    assert e.center == (0, 0, 0)
    assert e.major_axis == (1, 0, 0)
    assert e.minor_axis == (0, 1, 0)
    assert e.extrusion == (0, 0, 1)
    assert e.ratio == 1.0
    assert e.start_param == 0
    assert e.end_param == math.tau


def test_dxfattribs():
    e = ConstructionEllipse()
    attribs = e.dxfattribs()
    assert attribs["center"] == (0, 0, 0)
    assert attribs["major_axis"] == (1, 0, 0)
    assert "minor_axis" not in attribs
    assert attribs["extrusion"] == (0, 0, 1)
    assert attribs["ratio"] == 1.0
    assert attribs["start_param"] == 0
    assert attribs["end_param"] == math.tau


def test_get_start_and_end_vertex():
    ellipse = ConstructionEllipse(
        center=(1, 2, 3),
        major_axis=(4, 3, 0),
        extrusion=(0, 0, -1),
        ratio=0.7,
        start_param=math.pi / 2,
        end_param=math.pi,
    )

    start, end = list(
        ellipse.vertices(
            [
                ellipse.start_param,
                ellipse.end_param,
            ]
        )
    )
    # test values from BricsCAD
    assert start.isclose(Vec3(3.1, -0.8, 3), abs_tol=1e-6)
    assert end.isclose(Vec3(-3, -1, 3), abs_tol=1e-6)

    # for convenience, but vertices() is much more efficient:
    assert ellipse.start_point.isclose(Vec3(3.1, -0.8, 3), abs_tol=1e-6)
    assert ellipse.end_point.isclose(Vec3(-3, -1, 3), abs_tol=1e-6)


def test_from_arc():
    ellipse = ConstructionEllipse.from_arc(center=(2, 2, 2), radius=3)
    assert ellipse.center == (2, 2, 2)
    assert ellipse.major_axis == (3, 0, 0)
    assert ellipse.ratio == 1
    assert ellipse.start_param == 0
    assert math.isclose(ellipse.end_param, math.tau)


def test_swap_axis_full_ellipse():
    ellipse = ConstructionEllipse(
        major_axis=(5, 0, 0),
        ratio=2,
    )
    assert ellipse.minor_axis.isclose((0, 10, 0))

    ellipse.swap_axis()
    assert ellipse.ratio == 0.5
    assert ellipse.major_axis == (0, 10, 0)
    assert ellipse.minor_axis == (-5, 0, 0)
    assert ellipse.start_param == 0
    assert ellipse.end_param == math.pi * 2


def test_swap_axis_half_ellipse():
    ellipse = ConstructionEllipse(
        major_axis=(5, 0, 0),
        ratio=2,
        start_param=math.pi / 2.0,
        end_param=math.pi / 2.0 * 3.0,
    )
    assert ellipse.minor_axis.isclose((0, 10, 0))

    ellipse.swap_axis()
    assert ellipse.ratio == 0.5
    assert ellipse.major_axis == (0, 10, 0)
    assert ellipse.minor_axis == (-5, 0, 0)
    assert ellipse.start_param == 0
    assert ellipse.end_param == math.pi


def non_zero_random(limit=10):
    return random.uniform(0.001, limit) * random.choice((1, -1))


def test_swap_axis_arbitrary_params():
    random_tests_count = 100
    random.seed(0)

    for _ in range(random_tests_count):
        ellipse = ConstructionEllipse(
            # avoid (0, 0, 0) as major axis
            major_axis=(non_zero_random(), non_zero_random(), 0),
            ratio=2,
            start_param=random.uniform(0, math.tau),
            end_param=random.uniform(0, math.tau),
            extrusion=(0, 0, random.choice((1, -1))),
        )

        # Test if coordinates of start- and end point stay at the same location
        # before and after swapping axis.
        start_point = ellipse.start_point
        end_point = ellipse.end_point
        minor_axis = ellipse.minor_axis
        ellipse.swap_axis()
        assert ellipse.major_axis.isclose(minor_axis, abs_tol=1e-9)
        assert ellipse.start_point.isclose(start_point, abs_tol=1e-9)
        assert ellipse.end_point.isclose(end_point, abs_tol=1e-9)


def test_params():
    count = 9
    e = ConstructionEllipse(start_param=math.pi / 2, end_param=-math.pi / 2)
    params = list(e.params(count))
    expected = list(np.linspace(math.pi / 2, math.pi / 2.0 * 3.0, count))
    assert params == expected


def test_angle_to_param():
    random_tests_count = 100
    random.seed(0)

    angle = 1.23
    assert math.isclose(angle_to_param(1.0, angle), angle)

    angle = 1.23 + math.pi / 2
    assert math.isclose(angle_to_param(1.0, angle), angle)

    angle = 1.23 + math.pi
    assert math.isclose(angle_to_param(1.0, angle), angle)

    angle = 1.23 + 3 * math.pi / 2
    assert math.isclose(angle_to_param(1.0, angle), angle)

    angle = math.pi / 2 + 1e-15
    assert math.isclose(angle_to_param(1.0, angle), angle)

    for _ in range(random_tests_count):
        ratio = random.uniform(1e-6, 1)
        angle = random.uniform(0, math.tau)
        param = angle_to_param(ratio, angle)
        ellipse = ConstructionEllipse(
            # avoid (0, 0, 0) as major axis
            major_axis=(non_zero_random(), non_zero_random(), 0),
            ratio=ratio,
            start_param=0,
            end_param=param,
            extrusion=(0, 0, random.choice((1, -1))),
        )
        calculated_angle = ellipse.extrusion.angle_about(
            ellipse.major_axis, ellipse.end_point
        )
        calculated_angle_without_direction = ellipse.major_axis.angle_between(
            ellipse.end_point
        )
        assert math.isclose(calculated_angle, angle, abs_tol=1e-9)
        assert math.isclose(
            calculated_angle, calculated_angle_without_direction
        ) or math.isclose(
            math.tau - calculated_angle, calculated_angle_without_direction
        )


def test_vertices():
    e = ConstructionEllipse(
        center=(3, 3),
        major_axis=(2, 0),
        ratio=0.5,
        start_param=0,
        end_param=math.pi * 1.5,
    )
    params = list(e.params(7))
    result = [
        (5.0, 3.0, 0.0),
        (4.414213562373095, 3.7071067811865475, 0.0),
        (3.0, 4.0, 0.0),
        (1.585786437626905, 3.7071067811865475, 0.0),
        (1.0, 3.0, 0.0),
        (1.5857864376269046, 2.2928932188134525, 0.0),
        (3.0, 2.0, 0.0),
    ]
    for v, r in zip(e.vertices(params), result):
        assert v.isclose(r)

    v1, v2 = e.vertices([0, math.tau])
    assert v1.isclose(v2)


def test_tangents():
    e = ConstructionEllipse(
        center=(3, 3),
        major_axis=(2, 0),
        ratio=0.5,
        start_param=0,
        end_param=math.pi * 1.5,
    )
    params = list(e.params(7))
    result = [
        (0.0, 1.0, 0.0),
        (-0.894427190999916, 0.447213595499958, 0.0),
        (-1.0, 3.061616997868383e-17, 0.0),
        (-0.894427190999916, -0.4472135954999579, 0.0),
        (-2.4492935982947064e-16, -1.0, 0.0),
        (0.8944271909999159, -0.44721359549995804, 0.0),
        (1.0, 0.0, 0.0),
    ]
    for v, r in zip(e.tangents(params), result):
        assert v.isclose(r)


def test_params_from_vertices_random():
    center = Vec3.random(5)
    major_axis = Vec3.random(5)
    extrusion = Vec3.random()
    ratio = 0.75
    e = ConstructionEllipse(center, major_axis, extrusion, ratio)

    params = [random.uniform(0.0001, math.tau - 0.0001) for _ in range(20)]
    vertices = e.vertices(params)
    new_params = e.params_from_vertices(vertices)
    for expected, param in zip(params, new_params):
        assert math.isclose(expected, param)

    # This creates the same vertex as v1 and v2
    v1, v2 = e.vertices([0, math.tau])
    assert v1.isclose(v2)

    # This should create the same param for v1 and v2, but
    # floating point inaccuracy produces unpredictable results:
    p1, p2 = e.params_from_vertices((v1, v2))

    assert math.isclose(p1, 0, abs_tol=1e-9) or math.isclose(
        p1, math.tau, abs_tol=1e-9
    )
    assert math.isclose(p2, 0, abs_tol=1e-9) or math.isclose(
        p2, math.tau, abs_tol=1e-9
    )


def test_to_ocs():
    e = ConstructionEllipse().to_ocs()
    assert e.center == (0, 0)


@pytest.mark.parametrize(
    "s, e, distance, count",
    [
        # known tests from ARC
        (0, 180, 0.35, 3),
        (0, 180, 0.10, 5),
        (270, 90, 0.10, 5),  # start angle > end angle
        (90, -90, 0.10, 5),
        (0, 0, 0.10, 0),  # angle span 0 works but yields nothing
        (-45, -45, 0.10, 0),
    ],
)
def test_flattening(s, e, distance, count):
    ellipse = ConstructionEllipse(
        start_param=math.radians(s),
        end_param=math.radians(e),
    )
    assert len(list(ellipse.flattening(distance, segments=2))) == count


def test_flattening_ellipse():
    # Visually checked in BricsCAD:
    e = ConstructionEllipse(major_axis=(3, 0), ratio=0.25)
    assert len(list(e.flattening(0.1))) == 13
    assert len(list(e.flattening(0.01))) == 37



# Copyright (c) 2020-2023, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import (
    Optional,
    Iterator,
    Iterable,
    Any,
    Callable,
)
from typing_extensions import Self
from ezdxf.math import (
    Vec3,
    NULLVEC,
    OCS,
    Bezier3P,
    Bezier4P,
    Matrix44,
    has_clockwise_orientation,
    UVec,
    BoundingBox,
)

from .commands import (
    Command,
    LineTo,
    MoveTo,
    Curve3To,
    Curve4To,
    PathElement,
)

__all__ = ["Path"]

MAX_DISTANCE = 0.01
MIN_SEGMENTS = 4
G1_TOL = 1e-4
_slots = ("_vertices", "_start_index", "_commands", "_has_sub_paths", "_user_data")


class Path:
    __slots__ = _slots

    def __init__(self, start: UVec = NULLVEC):
        # stores all command vertices in a contiguous list:
        self._vertices: list[Vec3] = [Vec3(start)]
        # start index of each command
        self._start_index: list[int] = []
        self._commands: list[Command] = []
        self._has_sub_paths = False
        self._user_data: Any = None  # should be immutable data!

    @classmethod
    def from_vertices_and_commands(
        cls, vertices: list[Vec3], command_codes: list[Command], user_data: Any = None
    ) -> Self:
        """Create path instances from a list of vertices and a list of commands."""
        # Used for fast conversion from NumpyPath2d to Path.
        # This is "hacky" but also 8x faster than the correct way using only public
        # methods and properties.
        new_path = cls()
        if len(vertices) == 0:
            return new_path
        new_path._vertices = vertices
        new_path._commands = command_codes
        new_path._start_index = make_vertex_index(command_codes)
        new_path._has_sub_paths = any(cmd == Command.MOVE_TO for cmd in command_codes)
        new_path._user_data = user_data
        return new_path

    def transform(self, m: Matrix44) -> Self:
        """Returns a new transformed path.

        Args:
             m: transformation matrix of type :class:`~ezdxf.math.Matrix44`

        """
        new_path = self.clone()
        new_path._vertices = list(m.transform_vertices(self._vertices))
        return new_path

    def bbox(self) -> BoundingBox:
        """Returns the bounding box of all control vertices as
        :class:`~ezdxf.math.BoundingBox` instance.
        """
        return BoundingBox(self.control_vertices())

    def __len__(self) -> int:
        """Returns count of path elements."""
        return len(self._commands)

    def __getitem__(self, item) -> PathElement:
        """Returns the path element at given index, slicing is not supported."""
        if isinstance(item, slice):
            raise TypeError("slicing not supported")
        cmd = self._commands[item]
        index = self._start_index[item]
        vertices = self._vertices
        if cmd == Command.MOVE_TO:
            return MoveTo(vertices[index])
        if cmd == Command.LINE_TO:
            return LineTo(vertices[index])
        if cmd == Command.CURVE3_TO:  # end, ctrl
            return Curve3To(vertices[index + 1], vertices[index])
        if cmd == Command.CURVE4_TO:
            return Curve4To(  # end, ctrl1, ctrl2
                vertices[index + 2],
                vertices[index],
                vertices[index + 1],
            )
        raise ValueError(f"Invalid command: {cmd}")

    def __iter__(self) -> Iterator[PathElement]:
        return (self[i] for i in range(len(self._commands)))

    def commands(self) -> list[PathElement]:
        """Returns all path elements as list."""
        return list(self.__iter__())

    def __copy__(self) -> Self:
        """Returns a new copy of :class:`Path` with shared immutable data."""
        copy = self.__class__()
        # vertices itself are immutable - no copying required
        copy._vertices = self._vertices.copy()
        self._copy_properties(copy)
        return copy

    def _copy_properties(self, clone: Path) -> None:
        assert len(self._vertices) == len(clone._vertices)
        clone._commands = self._commands.copy()
        clone._start_index = self._start_index.copy()
        clone._has_sub_paths = self._has_sub_paths
        # copy by reference: user data should be immutable data!
        clone._user_data = self._user_data

    clone = __copy__

    @property
    def user_data(self) -> Any:
        """Attach arbitrary user data to a :class:`Path` object.
        The user data is copied by reference, no deep copy is applied
        therefore a mutable state is shared between copies.
        """
        return self._user_data

    @user_data.setter
    def user_data(self, data: Any):
        self._user_data = data

    @property
    def start(self) -> Vec3:
        """:class:`Path` start point, resetting the start point of an empty
        path is possible.
        """
        return self._vertices[0]

    @start.setter
    def start(self, location: UVec) -> None:
        if self._commands:
            raise ValueError("Requires an empty path.")
        else:
            self._vertices[0] = Vec3(location)

    @property
    def end(self) -> Vec3:
        """:class:`Path` end point."""
        return self._vertices[-1]

    def control_vertices(self) -> list[Vec3]:
        """Yields all path control vertices in consecutive order."""
        if self._commands:
            return list(self._vertices)
        return []

    def command_codes(self) -> list[int]:
        """Internal API."""
        return list(self._commands)

    @property
    def is_closed(self) -> bool:
        """Returns ``True`` if the start point is close to the end point."""
        vertices = self._vertices
        if len(vertices) > 1:
            return vertices[0].isclose(vertices[-1])
        return False

    @property
    def has_lines(self) -> bool:
        """Returns ``True`` if the path has any line segments."""
        return Command.LINE_TO in self._commands

    @property
    def has_curves(self) -> bool:
        """Returns ``True`` if the path has any curve segments."""
        return (
            Command.CURVE4_TO in self._commands or Command.CURVE3_TO in self._commands
        )

    @property
    def has_sub_paths(self) -> bool:
        """Returns ``True`` if the path is a :term:`Multi-Path` object that
        contains multiple sub-paths.

        """
        return self._has_sub_paths

    def has_clockwise_orientation(self) -> bool:
        """Returns ``True`` if 2D path has clockwise orientation, ignores
        z-axis of all control vertices.

        Raises:
            TypeError: can't detect orientation of a :term:`Multi-Path` object

        """
        if self.has_sub_paths:
            raise TypeError("can't detect orientation of a multi-path object")
        return has_clockwise_orientation(self._vertices)

    def append_path_element(self, cmd: PathElement) -> None:
        """Append a single path element."""
        t = cmd.type
        if t == Command.LINE_TO:
            self.line_to(cmd.end)
        elif t == Command.MOVE_TO:
            self.move_to(cmd.end)
        elif t == Command.CURVE3_TO:
            self.curve3_to(cmd.end, cmd.ctrl)  # type: ignore
        elif t == Command.CURVE4_TO:
            self.curve4_to(cmd.end, cmd.ctrl1, cmd.ctrl2)  # type: ignore
        else:
            raise ValueError(f"Invalid command: {t}")

    def line_to(self, location: UVec) -> None:
        """Add a line from actual path end point to `location`."""
        self._commands.append(Command.LINE_TO)
        self._start_index.append(len(self._vertices))
        self._vertices.append(Vec3(location))

    def move_to(self, location: UVec) -> None:
        """Start a new sub-path at `location`. This creates a gap between the
        current end-point and the start-point of the new sub-path. This converts
        the instance into a :term:`Multi-Path` object.

        If the :meth:`move_to` command is the first command, the start point of
        the path will be reset to `location`.

        """
        commands = self._commands
        if not commands:
            self._vertices[0] = Vec3(location)
            return
        self._has_sub_paths = True
        if commands[-1] == Command.MOVE_TO:
            # replace last move to command
            commands.pop()
            self._vertices.pop()
            self._start_index.pop()
        commands.append(Command.MOVE_TO)
        self._start_index.append(len(self._vertices))
        self._vertices.append(Vec3(location))

    def curve3_to(self, location: UVec, ctrl: UVec) -> None:
        """Add a quadratic Bèzier-curve from actual path end point to
        `location`, `ctrl` is the control point for the quadratic Bèzier-curve.
        """
        self._commands.append(Command.CURVE3_TO)
        self._start_index.append(len(self._vertices))
        self._vertices.extend((Vec3(ctrl), Vec3(location)))

    def curve4_to(self, location: UVec, ctrl1: UVec, ctrl2: UVec) -> None:
        """Add a cubic Bèzier-curve from actual path end point to `location`,
        `ctrl1` and `ctrl2` are the control points for the cubic Bèzier-curve.
        """
        self._commands.append(Command.CURVE4_TO)
        self._start_index.append(len(self._vertices))
        self._vertices.extend((Vec3(ctrl1), Vec3(ctrl2), Vec3(location)))

    def close(self) -> None:
        """Close path by adding a line segment from the end point to the start
        point.
        """
        if not self.is_closed:
            self.line_to(self.start)

    def close_sub_path(self) -> None:
        """Close last sub-path by adding a line segment from the end point to
        the start point of the last sub-path. Behaves like :meth:`close` for
        :term:`Single-Path` instances.
        """
        if self.has_sub_paths:
            start_point = self._start_of_last_sub_path()
            assert (
                start_point is not None
            ), "internal error: required MOVE_TO command not found"
            if not self.end.isclose(start_point):
                self.line_to(start_point)
        else:
            self.close()

    def _start_of_last_sub_path(self) -> Optional[Vec3]:
        move_to = Command.MOVE_TO
        commands = self._commands
        index = len(commands) - 1
        # The first command at index 0 is never MOVE_TO!
        while index > 0:
            if commands[index] == move_to:
                return self._vertices[self._start_index[index]]
            index -= 1
        return None

    def reversed(self) -> Self:
        """Returns a new :class:`Path` with reversed commands and control
        vertices.

        """
        path = self.clone()
        if not path._commands:
            return path
        if path._commands[-1] == Command.MOVE_TO:
            # The last move_to will become the first move_to.
            # A move_to as first command just moves the start point and can be
            # removed!
            # There are never two consecutive MOVE_TO commands in a Path!
            path._commands.pop()
            path._vertices.pop()
            path._start_index.pop()
            path._has_sub_paths = any(  # is still a multi-path?
                cmd == Command.MOVE_TO for cmd in path._commands
            )
        path._commands.reverse()
        path._vertices.reverse()
        path._start_index = make_vertex_index(path._commands)
        return path

    def clockwise(self) -> Self:
        """Returns new :class:`Path` in clockwise orientation.

        Raises:
            TypeError: can't detect orientation of a :term:`Multi-Path` object

        """
        if self.has_clockwise_orientation():
            return self.clone()
        else:
            return self.reversed()

    def counter_clockwise(self) -> Self:
        """Returns new :class:`Path` in counter-clockwise orientation.

        Raises:
            TypeError: can't detect orientation of a :term:`Multi-Path` object

        """

        if self.has_clockwise_orientation():
            return self.reversed()
        else:
            return self.clone()

    def approximate(self, segments: int = 20) -> Iterator[Vec3]:
        """Approximate path by vertices, `segments` is the count of
        approximation segments for each Bézier curve.

        Does not yield any vertices for empty paths, where only a start point
        is present!

        Approximation of :term:`Multi-Path` objects is possible, but gaps are
        indistinguishable from line segments.

        """

        def curve3(p0: Vec3, p1: Vec3, p2: Vec3) -> Iterator[Vec3]:
            return iter(Bezier3P((p0, p1, p2)).approximate(segments))

        def curve4(p0: Vec3, p1: Vec3, p2: Vec3, p3: Vec3) -> Iterator[Vec3]:
            return iter(Bezier4P((p0, p1, p2, p3)).approximate(segments))

        return self._approximate(curve3, curve4)

    def flattening(self, distance: float, segments: int = 4) -> Iterator[Vec3]:
        """Approximate path by vertices and use adaptive recursive flattening
        to approximate Bèzier curves. The argument `segments` is the
        minimum count of approximation segments for each curve, if the distance
        from the center of the approximation segment to the curve is bigger than
        `distance` the segment will be subdivided.

        Does not yield any vertices for empty paths, where only a start point
        is present!

        Flattening of :term:`Multi-Path` objects is possible, but gaps are
        indistinguishable from line segments.

        Args:
            distance: maximum distance from the center of the curve to the
                center of the line segment between two approximation points to
                determine if a segment should be subdivided.
            segments: minimum segment count per Bézier curve

        """

        def curve3(p0: Vec3, p1: Vec3, p2: Vec3) -> Iterator[Vec3]:
            if distance == 0.0:
                raise ValueError(f"invalid max distance: 0.0")
            return iter(Bezier3P((p0, p1, p2)).flattening(distance, segments))

        def curve4(p0: Vec3, p1: Vec3, p2: Vec3, p3: Vec3) -> Iterator[Vec3]:
            if distance == 0.0:
                raise ValueError(f"invalid max distance: 0.0")
            return iter(Bezier4P((p0, p1, p2, p3)).flattening(distance, segments))

        return self._approximate(curve3, curve4)

    def _approximate(self, curve3: Callable, curve4: Callable) -> Iterator[Vec3]:
        if not self._commands:
            return

        start = self._vertices[0]
        yield start

        vertices = self._vertices
        for si, cmd in zip(self._start_index, self._commands):
            if cmd == Command.LINE_TO or cmd == Command.MOVE_TO:
                end_location = vertices[si]
                yield end_location
            elif cmd == Command.CURVE3_TO:
                ctrl, end_location = vertices[si : si + 2]
                pts = curve3(start, ctrl, end_location)
                next(pts)  # skip first vertex
                yield from pts
            elif cmd == Command.CURVE4_TO:
                ctrl1, ctrl2, end_location = vertices[si : si + 3]
                pts = curve4(start, ctrl1, ctrl2, end_location)
                next(pts)  # skip first vertex
                yield from pts
            else:
                raise ValueError(f"Invalid command: {cmd}")
            start = end_location

    def to_wcs(self, ocs: OCS, elevation: float) -> None:
        """Transform path from given `ocs` to WCS coordinates inplace."""
        self._vertices = list(
            ocs.to_wcs(v.replace(z=float(elevation))) for v in self._vertices
        )

    def sub_paths(self) -> Iterator[Self]:
        """Yield all sub-paths as :term:`Single-Path` objects.

        It's safe to call :meth:`sub_paths` on any path-type:
        :term:`Single-Path`, :term:`Multi-Path` and :term:`Empty-Path`.

        """
        path = self.__class__(start=self.start)
        path._user_data = self._user_data
        move_to = Command.MOVE_TO
        for cmd in self.commands():
            if cmd.type == move_to:
                yield path
                path = self.__class__(start=cmd.end)
                path._user_data = self._user_data
            else:
                path.append_path_element(cmd)
        yield path

    def extend_multi_path(self, path: Path) -> None:
        """Extend the path by another path. The source path is automatically a
        :term:`Multi-Path` object, even if the previous end point matches the
        start point of the appended path. Ignores paths without any commands
        (empty paths).

        """
        if len(path):
            self.move_to(path.start)
            for cmd in path.commands():
                self.append_path_element(cmd)

    def append_path(self, path: Path) -> None:
        """Append another path to this path. Adds a :code:`self.line_to(path.start)`
        if the end of this path != the start of appended path.

        """
        if len(path) == 0:
            return  # do not append an empty path
        if self._commands:
            if not self.end.isclose(path.start):
                self.line_to(path.start)
        else:
            self.start = path.start
        for cmd in path.commands():
            self.append_path_element(cmd)


CMD_SIZE = {
    Command.MOVE_TO: 1,
    Command.LINE_TO: 1,
    Command.CURVE3_TO: 2,
    Command.CURVE4_TO: 3,
}


def make_vertex_index(command_codes: Iterable[Command]) -> list[int]:
    cmd_size = CMD_SIZE
    start: int = 1
    start_index: list[int] = []
    for code in command_codes:
        start_index.append(start)
        start += cmd_size[code]
    return start_index



# Copyright (c) 2020, Manfred Moitzi
# License: MIT License
import pytest
from ezdxf.math import (
    Plane,
    Vec3,
    X_AXIS,
    BoundingBox,
    Z_AXIS,
    split_polygon_by_plane,
)
from ezdxf.render import forms


def test_init():
    p = Plane(Vec3(1, 0, 0), 5)
    assert p.normal == (1, 0, 0)
    assert p.distance_from_origin == 5


def test_init_form_3_colinear_points():
    with pytest.raises(ValueError):
        Plane.from_3p(Vec3(1, 0, 0), Vec3(1, 0, 0), Vec3(1, 0, 0))


def test_init_form_3p():
    p = Plane.from_3p(Vec3(5, 0, 0), Vec3(5, 1, 5), Vec3(5, 0, 1))


def test_equal():
    p1 = Plane.from_vector((5, 0, 0))
    p2 = Plane.from_vector((5, 0, 0))
    assert p1 is not p2
    assert p1 == p1


def test_init_form_vector():
    p = Plane.from_vector((5, 0, 0))
    assert p.normal == (1, 0, 0)
    assert p.distance_from_origin == 5

    with pytest.raises(ValueError):
        Plane.from_vector((0, 0, 0))


def test_signed_distance_to():
    p = Plane.from_vector((5, 0, 0))
    assert p.signed_distance_to(Vec3(10, 0, 0)) == 5
    assert p.signed_distance_to(Vec3(0, 0, 0)) == -5


def test_distance_to():
    p = Plane.from_vector((5, 0, 0))
    assert p.distance_to(Vec3(10, 0, 0)) == 5
    assert p.distance_to(Vec3(0, 0, 0)) == 5


def test_is_coplanar():
    p = Plane.from_vector((5, 0, 0))
    assert p.is_coplanar_vertex(Vec3(5, 5, 0)) is True
    assert p.is_coplanar_vertex(Vec3(5, 0, 5)) is True


def test_is_coplanar_plane():
    p1 = Plane.from_vector((5, 0, 0))
    p2 = Plane.from_vector((-1, 0, 0))
    assert p1.is_coplanar_plane(p2) is True


class TestSplitConvexPolygon:
    def test_spit_horizontal_square(self):
        polygon = forms.square(center=True)
        plane = Plane(X_AXIS, 0)
        front, back = split_polygon_by_plane(polygon, plane)
        assert len(front) == 4
        assert len(back) == 4
        front_bbox = BoundingBox(front)
        assert front_bbox.extmin == (0, -0.5)
        assert front_bbox.extmax == (0.5, 0.5)
        back_bbox = BoundingBox(back)
        assert back_bbox.extmin == (-0.5, -0.5)
        assert back_bbox.extmax == (0, 0.5)

    def test_ignore_coplanar_square(self):
        polygon = forms.square(center=True)
        plane = Plane(Z_AXIS, 0)
        front, back = split_polygon_by_plane(polygon, plane, coplanar=False)
        assert len(front) == 0
        assert len(back) == 0

    def test_return_coplanar_square_front(self):
        polygon = forms.square(center=True)
        plane = Plane(Z_AXIS, 0)
        front, back = split_polygon_by_plane(polygon, plane, coplanar=True)
        assert len(front) == 4
        assert len(back) == 0

    def test_return_coplanar_square_back(self):
        polygon = forms.square(center=True)
        plane = Plane(-Z_AXIS, 0)
        front, back = split_polygon_by_plane(polygon, plane, coplanar=True)
        assert len(front) == 0
        assert len(back) == 4


class TestIntersectLine:
    @pytest.fixture(scope="class")
    def plane(self):
        return Plane(Z_AXIS, 5)

    def test_intersection_line(self, plane):
        ip = plane.intersect_line(Vec3(0, 0, 0), Vec3(0, 0, 10))
        assert ip.isclose((0, 0, 5))

    def test_line_above_plane(self, plane):
        ip = plane.intersect_line(Vec3(0, 0, 6), Vec3(0, 0, 10))
        assert ip is None

    def test_line_below_plane(self, plane):
        ip = plane.intersect_line(Vec3(0, 0, 0), Vec3(0, 0, 4))
        assert ip is None

    def test_colinear_start_point_intersection(self, plane):
        ip = plane.intersect_line(Vec3(0, 0, 5), Vec3(0, 0, 10))
        assert ip.isclose((0, 0, 5))

    def test_ignore_coplanar_start_point_intersection(self, plane):
        ip = plane.intersect_line(Vec3(0, 0, 5), Vec3(0, 0, 10), coplanar=False)
        assert ip is None

    def test_colinear_end_point_intersection(self, plane):
        ip = plane.intersect_line(Vec3(0, 0, 0), Vec3(0, 0, 5))
        assert ip.isclose((0, 0, 5))

    def test_ignore_coplanar_end_point_intersection(self, plane):
        ip = plane.intersect_line(Vec3(0, 0, 0), Vec3(0, 0, 5), coplanar=False)
        assert ip is None

    def test_coplanar_line_has_no_intersection(self, plane):
        ip = plane.intersect_line(Vec3(0, 0, 5), Vec3(1, 0, 5), coplanar=True)
        assert ip is None
        ip = plane.intersect_line(Vec3(0, 0, 5), Vec3(1, 0, 5), coplanar=False)
        assert ip is None


class TestIntersectRay:
    @pytest.fixture(scope="class")
    def plane(self):
        return Plane(Z_AXIS, 5)

    def test_intersection_line(self, plane):
        ip = plane.intersect_ray(Vec3(0, 0, 0), Vec3(0, 0, 10))
        assert ip.isclose((0, 0, 5))

    def test_coplanar_ray_has_no_intersection(self, plane):
        ip = plane.intersect_ray(Vec3(0, 0, 0), Vec3(1, 0, 0))
        assert ip is None

    def test_plane_parallel_to_yz(self):
        plane = Plane(Vec3(1, 0, 0), 3)
        ip = plane.intersect_ray(Vec3(0, 0, 2), Vec3(1, 0, 0))
        assert ip.isclose((3, 0, 2))


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2014-2019, Manfred Moitzi
# License: MIT License
import pytest
import os

import ezdxf

ZIPFILE = "read_zip_test.zip"


def get_zip_path(zip_file_name):
    test_path = os.path.dirname(__file__)
    zip_path = os.path.join(test_path, zip_file_name)
    return zip_path


@pytest.mark.skipif(
    not os.path.exists(get_zip_path(ZIPFILE)),
    reason="Skipped TestReadZip(): file '{}' not found.".format(
        get_zip_path(ZIPFILE)
    ),
)
def test_read_ac1009():
    dwg = ezdxf.readzip(get_zip_path(ZIPFILE), "AC1009.dxf")
    msp = dwg.modelspace()
    lines = msp.query("LINE")
    assert 255 == len(lines)



# cython: language_level=3
# Copyright (c) 2022-2024, Manfred Moitzi
# License: MIT License
from typing import Iterator, TYPE_CHECKING, Sequence
import cython
from cpython.mem cimport PyMem_Malloc, PyMem_Free
from .vector cimport Vec3, v3_isclose, v3_sub, v3_add, v3_mul, v3_magnitude

if TYPE_CHECKING:
    from ezdxf.math import UVec

__all__ = ["_LineTypeRenderer"]
LineSegment = tuple[Vec3, Vec3]

cdef extern from "constants.h":
    const double ABS_TOL
    const double REL_TOL


cdef class _LineTypeRenderer:
    cdef double *dashes
    cdef int dash_count
    cdef readonly bint is_solid
    cdef bint is_dash
    cdef int current_dash
    cdef double current_dash_length

    def __init__(self, dashes: Sequence[float]):
        cdef list _dashes = list(dashes)
        cdef int i

        self.dash_count = len(_dashes)
        self.dashes = <double *> PyMem_Malloc(self.dash_count * sizeof(double))
        for i in range(self.dash_count):
            self.dashes[i] = _dashes[i]

        self.is_solid = True
        self.is_dash = False
        self.current_dash = 0
        self.current_dash_length = 0.0

        if self.dash_count > 1:
            self.is_solid = False
            self.current_dash_length = self.dashes[0]
            self.is_dash = True

    def __dealloc__(self):
        PyMem_Free(self.dashes)

    def line_segment(self, start: UVec, end: UVec) -> Iterator[LineSegment]:
        cdef Vec3 v3_start = Vec3(start)
        cdef Vec3 v3_end = Vec3(end)
        cdef Vec3 segment_vec, segment_dir
        cdef double segment_length, dash_length
        cdef list dashes = []

        if self.is_solid or v3_isclose(v3_start, v3_end, REL_TOL, ABS_TOL):
            yield v3_start, v3_end
            return

        segment_vec = v3_sub(v3_end, v3_start)
        segment_length = v3_magnitude(segment_vec)
        with cython.cdivision:
            segment_dir = v3_mul(segment_vec, 1.0 / segment_length)  # normalize

        self._render_dashes(segment_length, dashes)
        for dash_length in dashes:
            v3_end = v3_add(v3_start, v3_mul(segment_dir, abs(dash_length)))
            if dash_length > 0:
                yield v3_start, v3_end
            v3_start = v3_end

    cdef _render_dashes(self, double length, list dashes):
        if length <= self.current_dash_length:
            self.current_dash_length -= length
            dashes.append(length if self.is_dash else -length)
            if self.current_dash_length < ABS_TOL:
                self._cycle_dashes()
        else:
            # Avoid deep recursions!
            while length > self.current_dash_length:
                length -= self.current_dash_length
                self._render_dashes(self.current_dash_length, dashes)
            if length > 0.0:
                self._render_dashes(length, dashes)

    cdef _cycle_dashes(self):
        with cython.cdivision:
            self.current_dash = (self.current_dash + 1) % self.dash_count
        self.current_dash_length = self.dashes[self.current_dash]
        self.is_dash = not self.is_dash



# Created: 25.01.2018
# Copyright (C) 2018-2019, Manfred Moitzi
# License: MIT License
import pytest
from ezdxf.lldxf.tagger import internal_tag_compiler
from ezdxf.lldxf.loader import load_dxf_structure
from ezdxf.lldxf.const import DXFStructureError


def test_loader():
    sections = load_dxf_structure(internal_tag_compiler(TEST_HEADER))
    assert len(sections) == 3
    header = sections["HEADER"]
    assert len(header) == 1  # header load_section has always only one entity
    header_entity = header[0]
    assert header_entity[0] == (0, "SECTION")
    assert header_entity[1] == (2, "HEADER")
    assert header_entity[2] == (9, "$ACADVER")
    assert header_entity[-1] == (3, "ANSI_1252")

    tables = sections["TABLES"]
    assert len(tables) == 1
    tables_header = tables[0]
    assert tables_header[0] == (0, "SECTION")
    assert tables_header[1] == (2, "TABLES")

    entities = sections["ENTITIES"]
    assert len(entities) == 1
    entities_header = entities[0]
    assert entities_header[0] == (0, "SECTION")
    assert entities_header[1] == (2, "ENTITIES")


def test_error_section():
    with pytest.raises(DXFStructureError):
        load_dxf_structure(internal_tag_compiler(SECTION_INVALID_NAME_TAG))

    with pytest.raises(DXFStructureError):
        load_dxf_structure(internal_tag_compiler(SECTION_NO_NAME_TAG))


def validator(text):
    tags = internal_tag_compiler(text)
    return load_dxf_structure(tags)


def test_valid_structure():
    sections = validator(
        "  0\nSECTION\n 2\nHEADER\n  0\nENDSEC\n  0\nSECTION\n  2\nCLASSES\n  0\nENDSEC\n  0\nEOF\n"
    )
    assert len(sections) == 2
    assert len(sections["HEADER"]) == 1  # ENDSEC is not present
    assert len(sections["CLASSES"]) == 1  # ENDSEC is not present


def test_eof_without_lineending():
    sections = validator(
        "  0\nSECTION\n 2\nHEADER\n  0\nENDSEC\n  0\nSECTION\n  2\nCLASSES\n  0\nENDSEC\n  0\nEOF"
    )
    assert len(sections) == 2
    assert len(sections["HEADER"]) == 1  # ENDSEC is not present
    assert len(sections["CLASSES"]) == 1  # ENDSEC is not present


def test_missing_eof():
    with pytest.raises(DXFStructureError):
        validator("999\ncomment")


def test_missing_endsec():
    with pytest.raises(DXFStructureError):
        validator(
            "  0\nSECTION\n 2\nHEADER\n  0\nSECTION\n  2\nCLASSES\n  0\nENDSEC\n  0\nEOF\n"
        )

    with pytest.raises(DXFStructureError):
        validator(
            "  0\nSECTION\n 2\nHEADER\n  0\nSECTION\n  2\nCLASSES\n  0\nEOF\n"
        )


def test_missing_endsec_and_eof():
    with pytest.raises(DXFStructureError):
        validator(
            "  0\nSECTION\n 2\nHEADER\n  0\nENDSEC\n  0\nSECTION\n  2\nCLASSES\n"
        )


def test_missing_section():
    with pytest.raises(DXFStructureError):
        validator(
            "  0\nENDSEC\n  0\nSECTION\n  2\nCLASSES\n  0\nENDSEC\n  0\nEOF\n"
        )


TEST_HEADER = """  0
SECTION
  2
HEADER
  9
$ACADVER
  1
AC1018
  9
$DWGCODEPAGE
  3
ANSI_1252
  0
ENDSEC
  0
SECTION
  2
TABLES
  0
ENDSEC
  0
SECTION
  2
ENTITIES
  0
ENDSEC
  0
EOF
"""

SECTION_INVALID_NAME_TAG = """  0
SECTION
  3
HEADER
  0
ENDSEC
"""

SECTION_NO_NAME_TAG = """  0
SECTION
  0
ENDSEC
"""



# Copyright (c) 2018-2022, Manfred Moitzi
# License: MIT License
from .arrows import ARROWS
from .r12spline import R12Spline
from .curves import Bezier, EulerSpiral, Spline, random_2d_path, random_3d_path
from .mesh import (
    MeshBuilder,
    MeshVertexMerger,
    MeshTransformer,
    MeshAverageVertexMerger,
    MeshDiagnose,
    FaceOrientationDetector,
    MeshBuilderError,
    NonManifoldMeshError,
    MultipleMeshesError,
    NodeMergingError,
    DegeneratedPathError,
)
from .trace import TraceBuilder
from .mleader import (
    MultiLeaderBuilder,
    MultiLeaderMTextBuilder,
    MultiLeaderBlockBuilder,
    ConnectionSide,
    HorizontalConnection,
    VerticalConnection,
    LeaderType,
    TextAlignment,
    BlockAlignment,
)



# Copyright (c) 2019-2022 Manfred Moitzi
# License: MIT License
import pytest
import math

from ezdxf.math import Vec3, Matrix44, arc_angle_span_deg
from ezdxf.entities.arc import Arc
from ezdxf.lldxf.const import DXF12, DXF2000
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text
from ezdxf.path import make_path

TEST_CLASS = Arc
TEST_TYPE = "ARC"

ENTITY_R12 = """0
ARC
5
0
8
0
10
0.0
20
0.0
30
0.0
40
1.0
50
0
51
360
"""

ENTITY_R2000 = """0
ARC
5
0
330
0
100
AcDbEntity
8
0
100
AcDbCircle
10
0.0
20
0.0
30
0.0
40
1.0
100
AcDbArc
50
0
51
360
"""


@pytest.fixture(params=[ENTITY_R12, ENTITY_R2000])
def entity(request):
    return TEST_CLASS.from_text(request.param)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert TEST_TYPE in ENTITY_CLASSES


def test_default_init():
    entity = TEST_CLASS()
    assert entity.dxftype() == TEST_TYPE
    assert entity.dxf.handle is None
    assert entity.dxf.owner is None


def test_default_new():
    entity = TEST_CLASS.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": "7",
            "center": (1, 2, 3),
            "radius": 2.5,
            "start_angle": 30,
            "end_angle": 290,
        },
    )
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 7
    assert entity.dxf.linetype == "BYLAYER"

    assert entity.dxf.center == (1, 2, 3)
    assert entity.dxf.center.x == 1, "is not Vec3 compatible"
    assert entity.dxf.center.y == 2, "is not Vec3 compatible"
    assert entity.dxf.center.z == 3, "is not Vec3 compatible"
    assert entity.dxf.radius == 2.5
    assert entity.dxf.start_angle == 30
    assert entity.dxf.end_angle == 290
    # can set DXF R2007 value
    entity.dxf.shadow_mode = 1
    assert entity.dxf.shadow_mode == 1
    assert entity.dxf.extrusion == (0.0, 0.0, 1.0)
    assert entity.dxf.hasattr("extrusion") is False, "just the default value"


def test_get_start_and_end_vertices_with_ocs():
    arc = TEST_CLASS.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "center": (1, 2, 3),
            "radius": 2.5,
            "start_angle": 90,
            "end_angle": 180,
            "extrusion": (0, 0, -1),
        },
    )
    # convenient properties
    assert arc.start_point.isclose(Vec3(-1, 4.5, -3), abs_tol=1e-6)
    assert arc.end_point.isclose(Vec3(1.5, 2, -3), abs_tol=1e-6)

    # more efficient method:
    start, end = list(arc.vertices([arc.dxf.start_angle, arc.dxf.end_angle]))
    assert start.isclose(Vec3(-1, 4.5, -3), abs_tol=1e-6)
    assert end.isclose(Vec3(1.5, 2, -3), abs_tol=1e-6)


def test_load_from_text(entity):
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 256, "default color is 256 (by layer)"
    assert entity.dxf.center == (0, 0, 0)
    assert entity.dxf.radius == 1
    assert entity.dxf.start_angle == 0
    assert entity.dxf.end_angle == 360


@pytest.mark.parametrize(
    "txt,ver", [(ENTITY_R2000, DXF2000), (ENTITY_R12, DXF12)]
)
def test_write_dxf(txt, ver):
    expected = basic_tags_from_text(txt)
    arc = TEST_CLASS.from_text(txt)
    collector = TagCollector(dxfversion=ver, optional=True)
    arc.export_dxf(collector)
    assert collector.tags == expected

    collector2 = TagCollector(dxfversion=ver, optional=False)
    arc.export_dxf(collector2)
    assert collector.has_all_tags(collector2)


def test_angles():
    arc = Arc.new(dxfattribs={"radius": 1, "start_angle": 30, "end_angle": 60})
    assert tuple(arc.angles(2)) == (30, 60)
    assert tuple(arc.angles(3)) == (30, 45, 60)

    arc.dxf.start_angle = 180
    arc.dxf.end_angle = 0
    assert tuple(arc.angles(2)) == (180, 0)
    assert tuple(arc.angles(3)) == (180, 270, 0)

    arc.dxf.start_angle = -90
    arc.dxf.end_angle = -180
    assert tuple(arc.angles(2)) == (270, 180)
    assert tuple(arc.angles(4)) == (270, 0, 90, 180)


def test_arc_default_ocs():
    arc = Arc.new(
        dxfattribs={
            "center": (2, 3, 4),
            "thickness": 2,
            "start_angle": 30,
            "end_angle": 60,
        }
    )
    # 1. rotation - 2. scaling - 3. translation
    m = Matrix44.chain(Matrix44.scale(2, 2, 3), Matrix44.translate(1, 1, 1))
    # default extrusion is (0, 0, 1), therefore scale(2, 2, ..) is a uniform scaling in the xy-play of the OCS
    arc.transform(m)

    assert arc.dxf.center == (5, 7, 13)
    assert arc.dxf.extrusion == (0, 0, 1)
    assert arc.dxf.thickness == 6
    assert math.isclose(arc.dxf.start_angle, 30, abs_tol=1e-9)
    assert math.isclose(arc.dxf.end_angle, 60, abs_tol=1e-9)

    arc.transform(Matrix44.z_rotate(math.radians(30)))
    assert math.isclose(arc.dxf.start_angle, 60, abs_tol=1e-9)
    assert math.isclose(arc.dxf.end_angle, 90, abs_tol=1e-9)


# See also ConstructionArc(): test suite 645 - test_flattening()
@pytest.mark.parametrize(
    "r, s, e, sagitta, count",
    [
        (1, 0, 180, 0.10, 5),
        (0, 0, 360, 0.10, 0),  # radius 0 works but yields nothing
        (-1, 0, 180, 0.35, 3),  # negative radius same as positive radius
        (1, 270, 90, 0.10, 5),  # start angle > end angle
    ],
)
def test_arc_flattening(r, s, e, sagitta, count):
    arc = Arc.new(
        dxfattribs={
            "radius": r,
            "start_angle": s,
            "end_angle": e,
        }
    )
    points = list(arc.flattening(sagitta))
    assert len(points) == count


def test_arc_3d_flattening_parallel_to_xy_plane():
    arc = Arc.new(
        dxfattribs={
            "center": Vec3(1, 1, 1),
            "radius": 1,
            "start_angle": 0,
            "end_angle": 90,
        }
    )
    points = list(arc.flattening(0.01))
    expected = [
        Vec3(2.0, 1.0, 1.0),
        Vec3(1.9659258262890682, 1.2588190451025207, 1.0),
        Vec3(1.8660254037844388, 1.5, 1.0),
        Vec3(1.7071067811865475, 1.7071067811865475, 1.0),
        Vec3(1.5, 1.8660254037844386, 1.0),
        Vec3(1.2588190451025207, 1.9659258262890682, 1.0),
        Vec3(1.0, 2.0, 1.0),
    ]
    for p, e in zip(points, expected):
        assert p.isclose(e)


def test_arc_3d_flattening():
    arc = Arc.new(
        dxfattribs={
            "center": Vec3(1, 1, 1),
            "radius": 1,
            "start_angle": 30,
            "end_angle": 120,
            "extrusion": Vec3(1, 0, 0),
        }
    )
    points = list(arc.flattening(0.01))
    expected = [
        Vec3(1.0, 1.866025403784, 1.5),
        Vec3(1.0, 1.707106781187, 1.707106781187),
        Vec3(1.0, 1.5, 1.866025403784),
        Vec3(1.0, 1.258819045103, 1.965925826289),
        Vec3(1.0, 1.0, 2.0),
        Vec3(1.0, 0.741180954897, 1.965925826289),
        Vec3(1.0, 0.5, 1.866025403784),
    ]
    for p, e in zip(points, expected):
        assert p.isclose(e)


def test_arc_flattening_returns_Vec3():
    arc = Arc.new(
        dxfattribs={
            "radius": 1,
            "start_angle": 0,
            "end_angle": 180,
        }
    )
    points = list(arc.flattening(0.1))
    assert isinstance(points[0], Vec3), "must return Vec3() instances"


def test_360_deg_arc_transformation():
    arc = Arc.new(
        dxfattribs={
            "radius": 1,
            "start_angle": 0,
            "end_angle": 360,
        }
    )
    count1 = len(list(make_path(arc).flattening(0.01)))
    arc.transform(Matrix44.translate(1, 0, 0))
    count2 = len(list(make_path(arc).flattening(0.01)))
    assert count1 == count2

    arc.transform(Matrix44.z_rotate(math.pi / 2))
    p = make_path(arc)
    count3 = len(list(p.flattening(0.01)))
    assert count1 == count3


@pytest.mark.parametrize("angle", [30, 180, 360])
@pytest.mark.parametrize(
    "reflexion", [(-1, 1, 1), (1, -1, 1), (1, 1, -1)], ids=["x", "y", "z"]
)
def test_30_deg_arc_reflexion(reflexion, angle):
    arc = Arc.new(
        dxfattribs={
            "radius": 1,
            "start_angle": 0,
            "end_angle": angle,
        }
    )
    x, y, z = reflexion
    arc.transform(Matrix44.scale(x, y, z))
    assert arc_angle_span_deg(
        arc.dxf.start_angle, arc.dxf.end_angle
    ) == pytest.approx(angle)


MALFORMED_ARC = """0
ARC
5
0
62
7
330
0
6
LT_EZDXF
8
LY_EZDXF
100
AcDbCircle
10
1.0
20
2.0
30
3.0
100
AcDbEntity
40
2.0
50
30
51
330
"""


def test_load_malformed_circle():
    arc = Arc.from_text(MALFORMED_ARC)
    assert arc.dxf.layer == "LY_EZDXF"
    assert arc.dxf.linetype == "LT_EZDXF"
    assert arc.dxf.color == 7
    assert arc.dxf.center.isclose((1, 2, 3))
    assert arc.dxf.radius == 2.0
    assert arc.dxf.start_angle == 30.0
    assert arc.dxf.end_angle == 330.0



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf.tools.text import scale_mtext_inline_commands


@pytest.mark.parametrize(
    "text",
    [
        r"word \H100; word",
        r"word \H100 word",
        r"\H100; word",
        r"\H100 word",
        r"word \H100;",
        r"word \H100",
        r"\H100;",
        r"\H100",
        r"\H100.000",
    ],
)
def test_scale_absolute_height_commands(text):
    result = scale_mtext_inline_commands(text, 2.0)
    assert r"\H200" in result


@pytest.mark.parametrize(
    "text",
    [
        r"word \H100; word",
        r"word \H100 word",
        r"\H100; word",
        r"\H100 word",
        r"word \H100;",
        r"word \H100",
        r"\H100;",
        r"\H100",
    ],
)
def test_scaling_preserves_content_structure(text):
    expected = text.replace("1", "2")
    result = scale_mtext_inline_commands(text, 2.0)
    assert result == expected


@pytest.mark.parametrize(
    "text",
    [
        r"word \H100x; word",
        r"word \H100x word",
        r"\H100x; word",
        r"\H100x word",
        r"word \H100x;",
        r"word \H100x",
        r"\H100x;",
        r"\H100x",
        r"\H100.000x",
    ],
)
def test_does_not_scale_relative_height_commands(text):
    result = scale_mtext_inline_commands(text, 2.0)
    assert result == text


@pytest.mark.parametrize(
    "text",
    [
        r"word \H; word",
        r"word \H word",
        r"word \H0 word",
        r"word \H0; word",
        r"\H; word",
        r"\H word",
        r"\H0 word",
        r"\H0; word",
        r"word \H;",
        r"word \H",
        r"word \H0",
        r"word \H0;",
    ],
)
def test_invalid_constructs_are_preserved(text):
    # ... it is not the domain of this function to validate the MTEXT content string:
    # garbage in, garbage out
    assert scale_mtext_inline_commands(text, 2.0) == text


def test_negative_factor():
    assert scale_mtext_inline_commands(r"\H100; word", -2.0) == r"\H200; word"


def test_empty_content():
    assert scale_mtext_inline_commands("", 2.0) == ""


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2020, Manfred Moitzi
# License: MIT License
import pytest
import ezdxf


@pytest.fixture(scope="module")
def doc():
    return ezdxf.new("R2000")


def test_layout_dict_is_not_hard_owner(doc):
    assert doc.rootdict["ACAD_LAYOUT"].is_hard_owner is False


def test_create_new_layout(doc):
    NAME = "CreateNewLayout"

    new_layout = doc.new_layout(NAME)
    with pytest.raises(ezdxf.DXFValueError):
        doc.new_layout(NAME), "Same name for another layout is invalid."

    assert NAME == new_layout.name, "Has incorrect name"
    assert (
        new_layout.dxf_layout in doc.objects
    ), "Is not stored in OBJECTS section"
    assert (
        new_layout.name in doc.rootdict["ACAD_LAYOUT"]
    ), "Not stored in LAYOUT dict"
    assert (
        new_layout.block_record.dxf.name in doc.block_records
    ), "Missing required BLOCK_RECORD"
    assert (
        new_layout.block_record.dxf.name in doc.blocks
    ), "Missing required BLockLayout"


@pytest.mark.parametrize("name", ["Model", "MODEL", "model"])
def test_reserved_model_space_name(doc, name):
    with pytest.raises(ezdxf.DXFValueError):
        doc.new_layout(name), f'Creating a layout "{name}" is not allowed'
    with pytest.raises(ezdxf.DXFValueError):
        doc.layouts.delete(name), f'Deleting "{name}" is not allowed'
    with pytest.raises(ezdxf.DXFValueError):
        doc.layouts.rename(
            name, "XXX"
        ), f'Renaming layout "{name}" is not allowed'
    with pytest.raises(ezdxf.DXFValueError):
        doc.layouts.rename(
            "XXX", name
        ), f'Renaming a layout to "{name}" is not allowed'


def test_case_insensitive_layout_names(doc):
    NAME = "CaseInSensitive"
    layout = doc.layouts.new(NAME)
    assert layout.name == NAME, "Expected original case sensitive name"
    assert (
        NAME.upper() in doc.layouts
    ), "Test for containment should be case insensitive"
    assert layout is doc.layouts.get(
        NAME.upper()
    ), "get() should be case insensitive"

    with pytest.raises(ezdxf.DXFValueError):
        doc.layouts.new(
            NAME.upper()
        ), "Can not create layout with existing name"

    NEW_NAME = "CaseInSensitiveNewName"
    doc.layouts.rename(NAME.upper(), NEW_NAME)
    assert layout.name == NEW_NAME, "Expected original case sensitive name"
    assert NAME not in doc.layouts, "Old name still exist in layouts table"
    assert NEW_NAME in doc.layouts, "New name not stored in layouts table"

    doc.layouts.delete(NEW_NAME.upper())
    assert (
        NEW_NAME not in doc.layouts
    ), "Deleted layout still stored in layouts table."


def test_create_and_delete_new_layout(doc):
    NAME = "CreateAndDeleteNewLayout"
    new_layout = doc.new_layout(NAME)
    assert NAME == new_layout.name, "Expected original case sensitive name"
    assert new_layout.dxf_layout in doc.objects
    assert new_layout.name in doc.rootdict["ACAD_LAYOUT"]
    assert new_layout.block_record.dxf.name in doc.block_records
    assert new_layout.block_record.dxf.name in doc.blocks

    layout_name = new_layout.name
    block_name = new_layout.block_record.dxf.name
    doc.delete_layout(layout_name)

    assert new_layout.dxf_layout not in doc.objects
    assert layout_name not in doc.rootdict["ACAD_LAYOUT"]
    assert new_layout.block_record.is_alive is False
    assert block_name not in doc.blocks


def test_set_active_layout(doc):
    NAME = "SetActiveLayout"
    new_layout = doc.new_layout(NAME)
    doc.layouts.set_active_layout(NAME)
    assert "*Paper_Space" == new_layout.block_record_name
    assert (
        new_layout.layout_key
        == doc.block_records.get("*Paper_Space").dxf.handle
    )


def test_rename_layout(doc):
    THE_NEW_NAME = "TheNewName"
    RENAME_ME = "RenameMe"
    layout = doc.layouts.new(RENAME_ME)
    doc.layouts.rename(RENAME_ME, THE_NEW_NAME)

    assert RENAME_ME not in doc.layouts, "Old name not removed"
    assert THE_NEW_NAME in doc.layouts, "New name not stored"
    assert layout.name == THE_NEW_NAME, "Layout not renamed"
    assert layout.dxf_layout.is_alive is True, "Destroyed DXFLayout object"
    assert layout.dxf_layout in doc.objects

    # Check if management dict ACAD_LAYOUT is maintained:
    dxf_layouts = doc.rootdict.get("ACAD_LAYOUT")
    assert RENAME_ME not in dxf_layouts
    assert THE_NEW_NAME in dxf_layouts


def test_rename_not_existing_layout(doc):
    with pytest.raises(ezdxf.DXFValueError):
        doc.layouts.rename("LayoutDoesNotExist", "XXX")



# Source package: "py3dbp" hosted on PyPI
# (c) Enzo Ruiz Pelaez
# https://github.com/enzoruiz/3dbinpacking
# License: MIT License
# Credits:
# - https://github.com/enzoruiz/3dbinpacking/blob/master/erick_dube_507-034.pdf
# - https://github.com/gedex/bp3d - implementation in Go
# - https://github.com/bom-d-van/binpacking - implementation in Go
#
# ezdxf add-on:
# License: MIT License
# (c) 2022, Manfred Moitzi:
# - refactoring
# - type annotations
# - adaptations:
#   - removing Decimal class usage
#   - utilizing ezdxf.math.BoundingBox for intersection checks
#   - removed non-distributing mode; copy packer and use different bins for each copy
# - additions:
#   - Item.get_transformation()
#   - shuffle_pack()
#   - pack_item_subset()
#   - DXF exporter for debugging
from __future__ import annotations
from typing import (
    Iterable,
    TYPE_CHECKING,
    TypeVar,
    Optional,
)
from enum import Enum, auto
import copy
import math
import random


from ezdxf.enums import TextEntityAlignment
from ezdxf.math import (
    Vec2,
    Vec3,
    UVec,
    BoundingBox,
    BoundingBox2d,
    AbstractBoundingBox,
    Matrix44,
)
from . import genetic_algorithm as ga

if TYPE_CHECKING:
    from ezdxf.eztypes import GenericLayoutType

__all__ = [
    "Item",
    "FlatItem",
    "Box",  # contains Item
    "Envelope",  # contains FlatItem
    "AbstractPacker",
    "Packer",
    "FlatPacker",
    "RotationType",
    "PickStrategy",
    "shuffle_pack",
    "pack_item_subset",
    "export_dxf",
]

UNLIMITED_WEIGHT = 1e99
T = TypeVar("T")
PI_2 = math.pi / 2


class RotationType(Enum):
    """Rotation type of an item:

    - W = width
    - H = height
    - D = depth

    """

    WHD = auto()
    HWD = auto()
    HDW = auto()
    DHW = auto()
    DWH = auto()
    WDH = auto()


class Axis(Enum):
    WIDTH = auto()
    HEIGHT = auto()
    DEPTH = auto()


class PickStrategy(Enum):
    """Order of how to pick items for placement."""

    SMALLER_FIRST = auto()
    BIGGER_FIRST = auto()
    SHUFFLE = auto()


START_POSITION: tuple[float, float, float] = (0, 0, 0)


class Item:
    """3D container item."""

    def __init__(
        self,
        payload,
        width: float,
        height: float,
        depth: float,
        weight: float = 0.0,
    ):
        self.payload = payload  # arbitrary associated Python object
        self.width = float(width)
        self.height = float(height)
        self.depth = float(depth)
        self.weight = float(weight)
        self._rotation_type = RotationType.WHD
        self._position = START_POSITION
        self._bbox: Optional[AbstractBoundingBox] = None

    def __str__(self):
        return (
            f"{str(self.payload)}({self.width}x{self.height}x{self.depth}, "
            f"weight: {self.weight}) pos({str(self.position)}) "
            f"rt({self.rotation_type}) vol({self.get_volume()})"
        )

    def copy(self):
        """Returns a copy, all copies have a reference to the same payload
        object.
        """
        return copy.copy(self)  # shallow copy

    @property
    def bbox(self) -> AbstractBoundingBox:
        if self._bbox is None:
            self._update_bbox()
        return self._bbox  # type: ignore

    def _update_bbox(self) -> None:
        v1 = Vec3(self._position)
        self._bbox = BoundingBox([v1, v1 + Vec3(self.get_dimension())])

    def _taint(self):
        self._bbox = None

    @property
    def rotation_type(self) -> RotationType:
        return self._rotation_type

    @rotation_type.setter
    def rotation_type(self, value: RotationType) -> None:
        self._rotation_type = value
        self._taint()

    @property
    def position(self) -> tuple[float, float, float]:
        """Returns the position of then lower left corner of the item in the
        container, the lower left corner is the origin (0, 0, 0).
        """
        return self._position

    @position.setter
    def position(self, value: tuple[float, float, float]) -> None:
        self._position = value
        self._taint()

    def get_volume(self) -> float:
        """Returns the volume of the item."""
        return self.width * self.height * self.depth

    def get_dimension(self) -> tuple[float, float, float]:
        """Returns the item dimension according the :attr:`rotation_type`."""
        rt = self.rotation_type
        if rt == RotationType.WHD:
            return self.width, self.height, self.depth
        elif rt == RotationType.HWD:
            return self.height, self.width, self.depth
        elif rt == RotationType.HDW:
            return self.height, self.depth, self.width
        elif rt == RotationType.DHW:
            return self.depth, self.height, self.width
        elif rt == RotationType.DWH:
            return self.depth, self.width, self.height
        elif rt == RotationType.WDH:
            return self.width, self.depth, self.height
        raise ValueError(rt)

    def get_transformation(self) -> Matrix44:
        """Returns the transformation matrix to transform the source entity
        located with the minimum extension corner of its bounding box in
        (0, 0, 0) to the final location including the required rotation.
        """
        x, y, z = self.position
        rt = self.rotation_type
        if rt == RotationType.WHD:  # width, height, depth
            return Matrix44.translate(x, y, z)
        if rt == RotationType.HWD:  # height, width, depth
            return Matrix44.z_rotate(PI_2) @ Matrix44.translate(
                x + self.height, y, z
            )
        if rt == RotationType.HDW:  # height, depth, width
            return Matrix44.xyz_rotate(PI_2, 0, PI_2) @ Matrix44.translate(
                x + self.height, y + self.depth, z
            )
        if rt == RotationType.DHW:  # depth, height, width
            return Matrix44.y_rotate(-PI_2) @ Matrix44.translate(
                x + self.depth, y, z
            )
        if rt == RotationType.DWH:  # depth, width, height
            return Matrix44.xyz_rotate(0, PI_2, PI_2) @ Matrix44.translate(
                x, y, z
            )
        if rt == RotationType.WDH:  # width, depth, height
            return Matrix44.x_rotate(PI_2) @ Matrix44.translate(
                x, y + self.depth, z
            )
        raise TypeError(rt)


class FlatItem(Item):
    """2D container item, inherited from :class:`Item`. Has a default depth of
    1.0.
    """

    def __init__(
        self,
        payload,
        width: float,
        height: float,
        weight: float = 0.0,
    ):
        super().__init__(payload, width, height, 1.0, weight)

    def _update_bbox(self) -> None:
        v1 = Vec2(self._position)
        self._bbox = BoundingBox2d([v1, v1 + Vec2(self.get_dimension())])

    def __str__(self):
        return (
            f"{str(self.payload)}({self.width}x{self.height}, "
            f"weight: {self.weight}) pos({str(self.position)}) "
            f"rt({self.rotation_type}) area({self.get_volume()})"
        )


class Bin:
    def __init__(
        self,
        name,
        width: float,
        height: float,
        depth: float,
        max_weight: float = UNLIMITED_WEIGHT,
    ):
        self.name = name
        self.width = float(width)
        if self.width <= 0.0:
            raise ValueError("invalid width")
        self.height = float(height)
        if self.height <= 0.0:
            raise ValueError("invalid height")
        self.depth = float(depth)
        if self.depth <= 0.0:
            raise ValueError("invalid depth")
        self.max_weight = float(max_weight)
        self.items: list[Item] = []

    def __len__(self):
        return len(self.items)

    def __iter__(self):
        return iter(self.items)

    def copy(self):
        """Returns a copy."""
        box = copy.copy(self)  # shallow copy
        box.items = list(self.items)
        return box

    def reset(self):
        """Reset the container to empty state."""
        self.items.clear()

    @property
    def is_empty(self) -> bool:
        return not len(self.items)

    def __str__(self) -> str:
        return (
            f"{str(self.name)}({self.width:.3f}x{self.height:.3f}x{self.depth:.3f}, "
            f"max_weight:{self.max_weight}) "
            f"vol({self.get_capacity():.3f})"
        )

    def put_item(self, item: Item, pivot: tuple[float, float, float]) -> bool:
        valid_item_position = item.position
        item.position = pivot
        x, y, z = pivot

        # Try all possible rotations:
        for rotation_type in self.rotations():
            item.rotation_type = rotation_type
            w, h, d = item.get_dimension()
            if self.width < x + w or self.height < y + h or self.depth < z + d:
                continue
            # new item fits inside the box at he current location and rotation:
            item_bbox = item.bbox
            if (
                not any(item_bbox.has_intersection(i.bbox) for i in self.items)
                and self.get_total_weight() + item.weight <= self.max_weight
            ):
                self.items.append(item)
                return True

        item.position = valid_item_position
        return False

    def get_capacity(self) -> float:
        """Returns the maximum fill volume of the bin."""
        return self.width * self.height * self.depth

    def get_total_weight(self) -> float:
        """Returns the total weight of all fitted items."""
        return sum(item.weight for item in self.items)

    def get_total_volume(self) -> float:
        """Returns the total volume of all fitted items."""
        return sum(item.get_volume() for item in self.items)

    def get_fill_ratio(self) -> float:
        """Return the fill ratio."""
        try:
            return self.get_total_volume() / self.get_capacity()
        except ZeroDivisionError:
            return 0.0

    def rotations(self) -> Iterable[RotationType]:
        return RotationType


class Box(Bin):
    """3D container inherited from :class:`Bin`."""

    pass


class Envelope(Bin):
    """2D container inherited from :class:`Bin`."""

    def __init__(
        self,
        name,
        width: float,
        height: float,
        max_weight: float = UNLIMITED_WEIGHT,
    ):
        super().__init__(name, width, height, 1.0, max_weight)

    def __str__(self) -> str:
        return (
            f"{str(self.name)}({self.width:.3f}x{self.height:.3f}, "
            f"max_weight:{self.max_weight}) "
            f"area({self.get_capacity():.3f})"
        )

    def rotations(self) -> Iterable[RotationType]:
        return RotationType.WHD, RotationType.HWD


def _smaller_first(bins: list, items: list) -> None:
    # SMALLER_FIRST is often very bad! Especially for many in small
    # amounts increasing sizes.
    bins.sort(key=lambda b: b.get_capacity())
    items.sort(key=lambda i: i.get_volume())


def _bigger_first(bins: list, items: list) -> None:
    # BIGGER_FIRST is the best strategy
    bins.sort(key=lambda b: b.get_capacity(), reverse=True)
    items.sort(key=lambda i: i.get_volume(), reverse=True)


def _shuffle(bins: list, items: list) -> None:
    # Better as SMALLER_FIRST
    random.shuffle(bins)
    random.shuffle(items)


PICK_STRATEGY = {
    PickStrategy.SMALLER_FIRST: _smaller_first,
    PickStrategy.BIGGER_FIRST: _bigger_first,
    PickStrategy.SHUFFLE: _shuffle,
}


class AbstractPacker:
    def __init__(self) -> None:
        self.bins: list[Bin] = []
        self.items: list[Item] = []
        self._init_state = True

    def copy(self):
        """Copy packer in init state to apply different pack strategies."""
        if self.is_packed:
            raise TypeError("cannot copy packed state")
        if not all(box.is_empty for box in self.bins):
            raise TypeError("bins contain data in unpacked state")
        packer = self.__class__()
        packer.bins = [box.copy() for box in self.bins]
        packer.items = [item.copy() for item in self.items]
        return packer

    @property
    def is_packed(self) -> bool:
        """Returns ``True`` if packer is packed, each packer can only be used
        once.
        """
        return not self._init_state

    @property
    def unfitted_items(self) -> list[Item]:  # just an alias
        """Returns the unfitted items."""
        return self.items

    def __str__(self) -> str:
        fill = ""
        if self.is_packed:
            fill = f", fill ratio: {self.get_fill_ratio()}"
        return f"{self.__class__.__name__}, {len(self.bins)} bins{fill}"

    def append_bin(self, box: Bin) -> None:
        """Append a container."""
        if self.is_packed:
            raise TypeError("cannot append bins to packed state")
        if not box.is_empty:
            raise TypeError("cannot append bins with content")
        self.bins.append(box)

    def append_item(self, item: Item) -> None:
        """Append a item."""
        if self.is_packed:
            raise TypeError("cannot append items to packed state")
        self.items.append(item)

    def get_fill_ratio(self) -> float:
        """Return the fill ratio of all bins."""
        total_capacity = self.get_capacity()
        if total_capacity == 0.0:
            return 0.0
        return self.get_total_volume() / total_capacity

    def get_capacity(self) -> float:
        """Returns the maximum fill volume of all bins."""
        return sum(box.get_capacity() for box in self.bins)

    def get_total_weight(self) -> float:
        """Returns the total weight of all fitted items in all bins."""
        return sum(box.get_total_weight() for box in self.bins)

    def get_total_volume(self) -> float:
        """Returns the total volume of all fitted items in all bins."""
        return sum(box.get_total_volume() for box in self.bins)

    def get_unfitted_volume(self) -> float:
        """Returns the total volume of all unfitted items."""
        return sum(item.get_volume() for item in self.items)

    def pack(self, pick=PickStrategy.BIGGER_FIRST) -> None:
        """Pack items into bins. Distributes all items across all bins."""
        PICK_STRATEGY[pick](self.bins, self.items)
        # items are removed from self.items while packing!
        self._pack(self.bins, list(self.items))
        # unfitted items remain in self.items

    def _pack(self, bins: Iterable[Bin], items: Iterable[Item]) -> None:
        """Pack items into bins, removes packed items from self.items!"""
        self._init_state = False
        for box in bins:
            for item in items:
                if self.pack_to_bin(box, item):
                    self.items.remove(item)
        # unfitted items remain in self.items

    def pack_to_bin(self, box: Bin, item: Item) -> bool:
        if not box.items:
            return box.put_item(item, START_POSITION)

        for axis in self._axis():
            for placed_item in box.items:
                w, h, d = placed_item.get_dimension()
                x, y, z = placed_item.position
                if axis == Axis.WIDTH:
                    pivot = (x + w, y, z)  # new item right of the placed item
                elif axis == Axis.HEIGHT:
                    pivot = (x, y + h, z)  # new item above of the placed item
                elif axis == Axis.DEPTH:
                    pivot = (x, y, z + d)  # new item on top of the placed item
                else:
                    raise TypeError(axis)
                if box.put_item(item, pivot):
                    return True
        return False

    @staticmethod
    def _axis() -> Iterable[Axis]:
        return Axis


def shuffle_pack(packer: AbstractPacker, attempts: int) -> AbstractPacker:
    """Random shuffle packing. Returns a new packer with the best packing result,
    the input packer is unchanged.
    """
    if attempts < 1:
        raise ValueError("expected attempts >= 1")
    best_ratio = 0.0
    best_packer = packer
    for _ in range(attempts):
        new_packer = packer.copy()
        new_packer.pack(PickStrategy.SHUFFLE)
        new_ratio = new_packer.get_fill_ratio()
        if new_ratio > best_ratio:
            best_ratio = new_ratio
            best_packer = new_packer
    return best_packer


def pack_item_subset(
    packer: AbstractPacker, picker: Iterable, strategy=PickStrategy.BIGGER_FIRST
) -> None:
    """Pack a subset of `packer.items`, which are chosen by an iterable
    yielding a True or False value for each item.
    """
    assert packer.is_packed is False
    chosen, rejects = get_item_subset(packer.items, picker)
    packer.items = chosen
    packer.pack(strategy)  # unfitted items remain in packer.items
    packer.items.extend(rejects)  # append rejects as unfitted items


def get_item_subset(
    items: list[Item], picker: Iterable
) -> tuple[list[Item], list[Item]]:
    """Returns a subset of `items`, where items are chosen by an iterable
    yielding a True or False value for each item.
    """
    chosen: list[Item] = []
    rejects: list[Item] = []
    count = 0
    for item, pick in zip(items, picker):
        count += 1
        if pick:
            chosen.append(item)
        else:
            rejects.append(item)

    if count < len(items):  # too few pick values given
        rejects.extend(items[count:])
    return chosen, rejects


class SubSetEvaluator(ga.Evaluator):
    def __init__(self, packer: AbstractPacker):
        self.packer = packer

    def evaluate(self, dna: ga.DNA) -> float:
        packer = self.run_packer(dna)
        return packer.get_fill_ratio()

    def run_packer(self, dna: ga.DNA) -> AbstractPacker:
        packer = self.packer.copy()
        pack_item_subset(packer, dna)
        return packer


class Packer(AbstractPacker):
    """3D Packer inherited from :class:`AbstractPacker`."""

    def add_bin(
        self,
        name: str,
        width: float,
        height: float,
        depth: float,
        max_weight: float = UNLIMITED_WEIGHT,
    ) -> Box:
        """Add a 3D :class:`Box` container."""
        box = Box(name, width, height, depth, max_weight)
        self.append_bin(box)
        return box

    def add_item(
        self,
        payload,
        width: float,
        height: float,
        depth: float,
        weight: float = 0.0,
    ) -> Item:
        """Add a 3D :class:`Item` to pack."""
        item = Item(payload, width, height, depth, weight)
        self.append_item(item)
        return item


class FlatPacker(AbstractPacker):
    """2D Packer inherited from :class:`AbstractPacker`. All containers and
    items used by this packer must have a depth of 1."""

    def add_bin(
        self,
        name: str,
        width: float,
        height: float,
        max_weight: float = UNLIMITED_WEIGHT,
    ) -> Envelope:
        """Add a 2D :class:`Envelope` container."""
        envelope = Envelope(name, width, height, max_weight)
        self.append_bin(envelope)
        return envelope

    def add_item(
        self,
        payload,
        width: float,
        height: float,
        weight: float = 0.0,
    ) -> Item:
        """Add a 2D :class:`FlatItem` to pack."""
        item = FlatItem(payload, width, height, weight)
        self.append_item(item)
        return item

    @staticmethod
    def _axis() -> Iterable[Axis]:
        return Axis.WIDTH, Axis.HEIGHT


def export_dxf(
    layout: "GenericLayoutType", bins: list[Bin], offset: UVec = (1, 0, 0)
) -> None:
    from ezdxf import colors

    offset_vec = Vec3(offset)
    start = Vec3()
    index = 0
    rgb = (colors.RED, colors.GREEN, colors.BLUE, colors.MAGENTA, colors.CYAN)
    for box in bins:
        m = Matrix44.translate(start.x, start.y, start.z)
        _add_frame(layout, box, "FRAME", m)
        for item in box.items:
            _add_mesh(layout, item, "ITEMS", rgb[index], m)
            index += 1
            if index >= len(rgb):
                index = 0
        start += offset_vec


def _add_frame(layout: "GenericLayoutType", box: Bin, layer: str, m: Matrix44):
    def add_line(v1, v2):
        line = layout.add_line(v1, v2, dxfattribs=attribs)
        line.transform(m)

    attribs = {"layer": layer}
    x0, y0, z0 = (0.0, 0.0, 0.0)
    x1 = float(box.width)
    y1 = float(box.height)
    z1 = float(box.depth)
    corners = [
        (x0, y0),
        (x1, y0),
        (x1, y1),
        (x0, y1),
        (x0, y0),
    ]
    for (sx, sy), (ex, ey) in zip(corners, corners[1:]):
        add_line((sx, sy, z0), (ex, ey, z0))
        add_line((sx, sy, z1), (ex, ey, z1))
    for x, y in corners[:-1]:
        add_line((x, y, z0), (x, y, z1))

    text = layout.add_text(box.name, height=0.25, dxfattribs=attribs)
    text.set_placement((x0 + 0.25, y1 - 0.5, z1))
    text.transform(m)


def _add_mesh(
    layout: "GenericLayoutType", item: Item, layer: str, color: int, m: Matrix44
):
    from ezdxf.render.forms import cube

    attribs = {
        "layer": layer,
        "color": color,
    }
    mesh = cube(center=False)
    sx, sy, sz = item.get_dimension()
    mesh.scale(sx, sy, sz)
    x, y, z = item.position
    mesh.translate(x, y, z)
    mesh.render_polyface(layout, attribs, matrix=m)
    text = layout.add_text(
        str(item.payload), height=0.25, dxfattribs={"layer": "TEXT"}
    )
    if sy > sx:
        text.dxf.rotation = 90
        align = TextEntityAlignment.TOP_LEFT
    else:
        align = TextEntityAlignment.BOTTOM_LEFT
    text.set_placement((x + 0.25, y + 0.25, z + sz), align=align)
    text.transform(m)



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License

import pytest
from xml.etree import ElementTree as ET

from ezdxf.math import Vec2
from ezdxf.npshapes import NumpyPoints2d
from ezdxf.addons.drawing import svg, layout
from ezdxf.addons.drawing.properties import BackendProperties


class TestStyles:
    @pytest.fixture
    def xml(self):
        return ET.Element("defs")

    def test_create_stroke_style(self, xml):
        styles = svg.Styles(xml)
        cls = styles.get_class(stroke="black", stroke_width=10)
        assert cls == "C1"
        assert len(xml) == 1

        # repeat
        cls = styles.get_class(stroke="black", stroke_width=10)
        assert cls == "C1"
        assert len(xml) == 1

    def test_style_string(self, xml):
        styles = svg.Styles(xml)
        styles.get_class(stroke="black", stroke_width=10, stroke_opacity=0.5)

        string = ET.tostring(xml[0], encoding="unicode")
        assert (
            string
            == "<style>.C1 {stroke: black; stroke-width: 10; stroke-opacity: 0.500; "
            "fill: none; fill-opacity: 1.000;}</style>"
        )


class TestStrokeWidthMapping:
    def test_min_stroke_width(self):
        assert svg.map_lineweight_to_stroke_width(0.0, 10, 100) == 10

    def test_max_stroke_width(self):
        assert svg.map_lineweight_to_stroke_width(3.0, 10, 100) == 100

    def test_interpolation(self):
        # min_lineweight = 0.05 defined by DXF
        # max_lineweight = 2.11 defined by DXF
        assert svg.map_lineweight_to_stroke_width(0.25, 5, 211) == 25
        assert svg.map_lineweight_to_stroke_width(0.5, 5, 211) == 50
        assert svg.map_lineweight_to_stroke_width(1.0, 5, 211) == 100
        assert svg.map_lineweight_to_stroke_width(0.05, 5, 211) == 5
        assert svg.map_lineweight_to_stroke_width(2.11, 5, 211) == 211


NS = {
    "http://www.w3.org/2000/svg": "",
}


class TestSVGBackend:
    @pytest.fixture
    def backend(self):
        backend_ = svg.SVGBackend()
        properties = BackendProperties(color="#ff0000", lineweight=0.25)
        points = NumpyPoints2d(Vec2.list([(0, 0), (100, 0), (100, 100), (0, 100)]))
        backend_.draw_filled_polygon(points, properties)
        return backend_

    def test_get_xml_root_element(self, backend):
        xml = backend.get_xml_root_element(layout.Page(400, 300))
        assert xml.tag.endswith("svg") is True
        assert xml.attrib["width"] == "400mm"
        assert xml.attrib["height"] == "300mm"
        assert xml.attrib["viewBox"] == "0 0 1000000 750000"

    def test_get_string(self, backend):
        result = backend.get_string(layout.Page(400, 300))
        assert isinstance(result, str) is True
        xml = ET.fromstring(result)
        assert xml.tag.endswith("svg") is True
        assert xml.attrib["width"] == "400mm"
        assert xml.attrib["height"] == "300mm"
        assert xml.attrib["viewBox"] == "0 0 1000000 750000"


def test_empty_page():
    backend_ = svg.SVGBackend()
    backend_.draw_point(Vec2(0, 0), BackendProperties())
    result = backend_.get_string(layout.Page(0, 0), xml_declaration=False)
    assert result == "<svg />"


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License

import pytest
import ezdxf
from ezdxf.layouts import VirtualLayout
from ezdxf import bbox, disassemble
from ezdxf.render.forms import square, translate


@pytest.fixture(scope="module")
def points1():
    lay = VirtualLayout()
    lay.add_point((-1, -2, -3))
    lay.add_point((4, 5, 6))
    return lay


def test_extend(points1):
    box = bbox.extents(points1)
    assert box.extmin == (-1, -2, -3)
    assert box.extmax == (4, 5, 6)


def test_line():
    lay = VirtualLayout()
    lay.add_line((0, 0), (10, 0))
    lay.add_line((0, 1), (10, 1))
    box = bbox.extents(lay)
    assert box.extmin == (0, 0, 0)
    assert box.extmax == (10, 1, 0)


def solid_entities():
    lay = VirtualLayout()
    lay.add_solid(translate(square(1), (-10, -10)))
    lay.add_solid(translate(square(1), (10, 10)))
    return lay


def solid_blockrefs():
    doc = ezdxf.new()
    blk = doc.blocks.new("Solid")
    blk.add_solid(square(1))
    msp = doc.modelspace()
    msp.add_blockref("Solid", (-10, -10))
    msp.add_blockref("Solid", (10, 10))
    return msp


def test_empty_cache():
    assert bbox.Cache().has_data is False


@pytest.mark.parametrize(
    "solids",
    [solid_entities(), solid_blockrefs()],
    ids=["entities", "blockrefs"],
)
@pytest.mark.parametrize("func", [bbox.multi_flat, bbox.multi_recursive])
def test_multi_boxes(solids, func):
    box = list(func(solids))
    assert box[0].extmin == (-10, -10)
    assert box[0].extmax == (-9, -9)
    assert box[1].extmin == (10, 10)
    assert box[1].extmax == (11, 11)


def test_cache_usage_without_handles(points1):
    # Entities in VirtualLayouts have no handles:
    cache = bbox.Cache()
    for _ in range(10):
        bbox.extents(points1, cache=cache)
    assert cache.hits == 0


def test_cache_usage_with_uuids(points1):
    # Entities in VirtualLayouts have no handles:
    cache = bbox.Cache(uuid=True)
    for _ in range(10):
        bbox.extents(points1, cache=cache)
    assert cache.hits == 18


@pytest.fixture(scope="module")
def msp_solids():
    return solid_blockrefs()


@pytest.mark.parametrize("func", [bbox.multi_flat, bbox.extents])
def test_cache_usage_for_flat_multi_boxes(msp_solids, func):
    cache = bbox.Cache()
    for _ in range(10):
        list(func(msp_solids, cache=cache))

    # This works because flat processing has not to yield bounding boxes for
    # sub entities, caching top level bounding boxes works well.
    assert cache.misses == 2 + 2  # first 2xINSERT, 2xSOLID in INSERT
    assert cache.hits == 9 * 2  # 9 x 2xINSERT


def test_cache_usage_for_recreation_on_the_fly(msp_solids):
    cache = bbox.Cache(uuid=True)
    for _ in range(10):
        list(bbox.multi_recursive(msp_solids, cache=cache))

    # This does not work well, because recursive processing has to yield the
    # bounding boxes for all sub entities. These sub entities are created on the
    # fly for every call and do not have a handle and always gets a new UUID,
    # which is the meaning of an UUID.
    # The INSERT entity, which has a handle, gets no own bounding box and is
    # therefore not cached.
    assert cache.misses == 20
    assert cache.hits == 0


@pytest.mark.parametrize(
    "func", [bbox.multi_flat, bbox.extents, bbox.multi_recursive]
)
def test_cache_usage_for_reused_virtual_entities(msp_solids, func):
    cache = bbox.Cache()
    # Create flat entity structure by yourself, so that virtual entities are
    # only created once:
    entities = list(disassemble.recursive_decompose(msp_solids))
    for _ in range(10):
        list(func(entities, cache=cache))

    # This does not work well by "handle only" usage, because 'entities' contains
    # virtual entities which have no handle and therefore are not cached:
    # multi_flat and extents, have a second access stage and triggers 2x20 cache
    # misses but this is just a cache access issue, this does not trigger 40
    # bounding box calculations!
    # multi_recursive is the lowest level and has only 20 cache misses.
    assert cache.misses in (20, 40)  # virtual entities do not have handles
    assert cache.hits == 0  # parent INSERT bbox is not calculated and cached


@pytest.mark.parametrize(
    "func", [bbox.multi_flat, bbox.extents, bbox.multi_recursive]
)
def test_cache_usage_with_uuids_for_reused_virtual_entities(msp_solids, func):
    cache = bbox.Cache(uuid=True)
    # Create flat entity structure by yourself, so that virtual entities are
    # only created once:
    entities = list(disassemble.recursive_decompose(msp_solids))
    for _ in range(10):
        list(func(entities, cache=cache))

    # This works, because virtual entities are cached by UUID
    # multi_flat and extents, have a second access stage: 2x2 misses, but
    # triggers only 2 bounding box calculations.
    # multi_recursive is the lowest level and has only 2 cache misses.
    assert cache.misses in (2, 4)
    assert cache.hits == 18


@pytest.fixture
def circle():
    lay = VirtualLayout()
    lay.add_circle((0, 0), radius=100)
    return lay


def test_fast_bounding_box_calculation(circle):
    box = bbox.extents(circle, fast=True)
    assert box.extmin == (-100, -100)
    assert box.extmax == (+100, +100)


def test_precise_bounding_box_calculation(circle):
    box = bbox.extents(circle, fast=False)
    assert box.extmin == (-100, -100)
    assert box.extmax == (+100, +100)


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2019-2020 Manfred Moitzi
# License: MIT License
import pytest
import ezdxf

from ezdxf.entities.polyline import Polyline
from ezdxf.lldxf.const import DXF12, DXF2000
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text
from ezdxf.math import Vec3


ENTITY_R12 = """0
POLYLINE
5
0
8
0
66
1
10
0.0
20
0.0
30
0.0
70
0
"""

ENTITY_R2000 = """0
POLYLINE
5
0
330
0
100
AcDbEntity
8
0
100
AcDb2dPolyline
66
1
10
0.0
20
0.0
30
0.0
70
0
"""


@pytest.fixture(params=[ENTITY_R12, ENTITY_R2000])
def entity(request):
    return Polyline.from_text(request.param)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert "POLYLINE" in ENTITY_CLASSES


def test_default_constructor():
    entity = Polyline()
    assert entity.dxftype() == "POLYLINE"
    assert entity.is_virtual is True
    assert entity.seqend is None, "SEQEND must not exist"


def test_default_new():
    entity = Polyline.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": "7",
        },
    )
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 7
    assert entity.dxf.linetype == "BYLAYER"
    # can set DXF R2007 value
    entity.dxf.shadow_mode = 1
    assert entity.dxf.shadow_mode == 1


def test_load_from_text(entity):
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 256, "default color is 256 (by layer)"


def test_polyline_3d():
    polyline = Polyline.new(dxfattribs={"flags": Polyline.POLYLINE_3D})
    collector = TagCollector(dxfversion="R2000", optional=True)
    polyline.export_dxf(collector)
    assert (100, "AcDb3dPolyline") == collector.tags[5]


def test_poly_face_mesh():
    polyline = Polyline.new(dxfattribs={"flags": Polyline.POLYFACE})
    collector = TagCollector(dxfversion="R2000", optional=True)
    polyline.export_dxf(collector)
    assert (100, "AcDbPolyFaceMesh") == collector.tags[5]


def test_polygon_mesh():
    polyline = Polyline.new(dxfattribs={"flags": Polyline.POLYMESH})
    collector = TagCollector(dxfversion="R2000", optional=True)
    polyline.export_dxf(collector)
    assert (100, "AcDbPolygonMesh") == collector.tags[5]


def test_copy_polyline():
    doc = ezdxf.new()
    msp = doc.modelspace()
    polyline = msp.add_polyline2d([(1, 2), (7, 8), (4, 3)])
    assert isinstance(polyline, Polyline)
    assert len(polyline) == 3

    copy = polyline.copy()
    assert isinstance(polyline, Polyline)
    assert len(copy) == 3
    assert list(polyline.points()) == list(copy.points())
    assert polyline.vertices is not copy.vertices
    assert polyline.vertices[0] is not copy.vertices[0]
    copy.append_vertices([(9, 9)])
    assert len(polyline) == 3
    assert len(copy) == 4

    assert copy not in msp, "is not assigned to modelspace"
    # but only one polyline is stored
    assert len(msp) == 1
    msp.add_entity(copy)
    assert len(msp) == 2
    assert polyline.dxf.handle != copy.dxf.handle
    assert polyline.dxf.owner == copy.dxf.owner
    for vertex in copy.vertices:
        assert (
            vertex.dxf.owner == copy.dxf.handle
        ), "POLYLINE is owner of the VERTEX entities"


@pytest.mark.parametrize("txt,ver", [(ENTITY_R2000, DXF2000), (ENTITY_R12, DXF12)])
def test_write_dxf(txt, ver):
    expected = basic_tags_from_text(txt)
    polyline = Polyline.from_text(txt)
    collector = TagCollector(dxfversion=ver, optional=True)
    polyline.export_dxf(collector)
    assert collector.tags == expected

    collector2 = TagCollector(dxfversion=ver, optional=False)
    polyline.export_dxf(collector2)
    assert collector.has_all_tags(collector2)


def test_polyline2d_transform_interface():
    pline = Polyline()
    pline.append_vertices([(0, 0, 0), (2, 0, 0), (1, 1, 0)])
    pline.translate(1, 1, 1)
    vertices = list(v.dxf.location for v in pline.vertices)
    assert pline.is_2d_polyline is True
    assert vertices[0] == (1, 1, 1)
    assert vertices[1] == (3, 1, 1)
    assert vertices[2] == (2, 2, 1)
    assert pline.dxf.elevation == (0, 0, 1)
    assert Vec3(0, 0, 1).isclose(pline.dxf.extrusion)


def test_polyline3d_transform_interface():
    pline = Polyline.new(dxfattribs={"flags": 8})
    pline.append_vertices([(0, 0, 0), (2, 0, 0), (1, 1, 0)])
    pline.translate(1, 1, 1)
    vertices = list(v.dxf.location for v in pline.vertices)
    assert pline.is_3d_polyline is True
    assert vertices[0] == (1, 1, 1)
    assert vertices[1] == (3, 1, 1)
    assert vertices[2] == (2, 2, 1)


def test_2d_polyline_has_default_width():
    assert Polyline().has_width is False
    assert Polyline.new(dxfattribs={"default_start_width": 0.1}).has_width is True
    assert Polyline.new(dxfattribs={"default_end_width": 0.1}).has_width is True


def test_2d_polyline_has_any_start_width():
    pline = Polyline()
    pline.append_formatted_vertices([(0.0, 0.0, 0.1)], format="xys")
    assert pline.has_width is True


def test_2d_polyline_has_any_end_width():
    pline = Polyline()
    pline.append_formatted_vertices([(0.0, 0.0, 0.1)], format="xye")
    assert pline.has_width is True


def test_2d_polyline_has_any_arc():
    pline = Polyline()
    assert pline.has_arc is False
    pline.append_formatted_vertices([(0.0, 0.0, 1.0)], format="xyb")
    assert pline.has_arc is True


MALFORMED_POLYLINE = """0
POLYLINE
5
0
62
7
330
0
6
LT_EZDXF
8
LY_EZDXF
100
AcDbEntity
66
1
10
0.0
20
0.0
30
99.0
70
0
100
AcDb2dPolyline
"""


def test_malformed_polyline():
    entity = Polyline.from_text(MALFORMED_POLYLINE)
    assert entity.dxf.layer == "LY_EZDXF"
    assert entity.dxf.linetype == "LT_EZDXF"
    assert entity.dxf.color == 7
    assert entity.dxf.elevation.isclose((0, 0, 99))



#  Copyright (c) 2021-2024, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf.math import Vec3
from ezdxf.entities import (
    BoundaryPaths,
    BoundaryPathType,
    EdgeType,
)


def test_add_polyline_path():
    paths = BoundaryPaths()
    new_path = paths.add_polyline_path([(0, 0), (0, 1), (1, 1), (1, 0)])
    assert new_path.type == BoundaryPathType.POLYLINE, "expected polyline path"
    assert 4 == len(new_path.vertices), "invalid vertex count"

    # now check the created entity
    assert 1 == len(paths)
    path = paths[0]
    assert path.type == BoundaryPathType.POLYLINE, "invalid path type"
    assert 4 == len(path.vertices), "invalid vertex count"
    # vertex format: x, y, bulge_value
    assert (0, 0, 0) == path.vertices[0], "invalid first vertex"
    assert (1, 0, 0) == path.vertices[3], "invalid last vertex"
    assert path.is_closed == 1


def test_add_edge_path():
    paths = BoundaryPaths()
    path = paths.add_edge_path()
    assert path.type == BoundaryPathType.EDGE, "expected edge path type"
    assert paths.has_edge_paths is True

    path.add_line((0, 0), (10, 0))
    path.add_arc((10, 5), radius=5, start_angle=270, end_angle=450, ccw=True)
    path.add_ellipse(
        (5, 10), major_axis=(5, 0), ratio=0.2, start_angle=0, end_angle=180
    )
    path.add_line((10, 0), (0, 0))
    # exit with statement and create DXF tags

    path = paths[-1]
    edge = path.edges[0]
    assert edge.type == EdgeType.LINE, "invalid edge type for 1. edge"
    assert (0, 0) == edge.start
    assert (10, 0) == edge.end

    edge = path.edges[1]
    assert edge.type == EdgeType.ARC, "invalid edge type for 2. edge"
    assert (10, 5) == edge.center
    assert 5 == edge.radius
    assert 270 == edge.start_angle
    assert 450 == edge.end_angle
    assert edge.ccw is True

    edge = path.edges[2]
    assert edge.type == EdgeType.ELLIPSE, "invalid edge type for 3. edge"
    assert (5, 10) == edge.center
    assert (5, 0) == edge.major_axis
    assert 0.2 == edge.ratio
    assert 0 == edge.start_angle
    assert 180 == edge.end_angle
    assert edge.ccw is True

    edge = path.edges[3]
    assert edge.type == EdgeType.LINE, "invalid edge type for 4. edge"
    assert (10, 0) == edge.start
    assert (0, 0) == edge.end


def test_polyline_path_with_arcs_to_edge_path():
    paths = BoundaryPaths()
    # <-- 2 ----
    #  \        \
    #   3 cw     1 ccw
    #  /        /
    # --- 0 --->
    paths.add_polyline_path([(0, 0), (10, 0, 1), (10, 4), (0, 4, -1), (0, 0)])
    paths.polyline_to_edge_paths()
    path = paths[0]
    assert (
        path.type == BoundaryPathType.EDGE
    ), "polyline path not converted to edge path"

    line = path.edges[0]
    assert line.type == EdgeType.LINE
    assert line.start == (0, 0)
    assert line.end == (10, 0)

    # counter-clockwise arc edge
    arc = path.edges[1]
    assert arc.type == EdgeType.ARC
    assert arc.center.isclose((10, 2))
    assert arc.start_angle == pytest.approx(270)
    assert arc.end_angle == pytest.approx(90)
    assert arc.ccw is True

    line = path.edges[2]
    assert line.type == EdgeType.LINE
    assert line.start == (10, 4)
    assert line.end == (0, 4)

    # clockwise arc edge
    arc = path.edges[3]
    assert arc.type == EdgeType.ARC
    assert arc.center.isclose((0, 2))
    # angles are in counter-clockwise order
    assert arc.start_angle == pytest.approx(270)
    assert arc.end_angle == pytest.approx(90)
    # but ccw flag forces clockwise DXF export
    assert arc.ccw is False


def test_arc_to_ellipse_edges():
    paths = BoundaryPaths()
    paths.add_polyline_path(
        [(0, 0, 1), (10, 0), (10, 10, -0.5), (0, 10)], is_closed=True
    )

    paths.polyline_to_edge_paths()
    path = paths[0]
    assert (
        path.type == BoundaryPathType.EDGE
    ), "polyline path not converted to edge path"

    paths.arc_edges_to_ellipse_edges()

    edge = path.edges[0]
    assert edge.type == EdgeType.ELLIPSE
    assert edge.center == (5, 0)
    assert edge.major_axis == (5, 0)
    assert edge.ratio == 1.0

    edge = path.edges[1]
    assert edge.type == EdgeType.LINE
    assert edge.start == (10, 0)
    assert edge.end == (10, 10)

    edge = path.edges[2]
    assert edge.type == EdgeType.ELLIPSE
    assert edge.ratio == 1.0

    edge = path.edges[3]
    assert edge.type == EdgeType.LINE
    assert edge.start == (0, 10)
    assert edge.end == (0, 0)


def test_ellipse_edges_to_spline_edges():
    paths = BoundaryPaths()
    paths.add_polyline_path(
        [(0, 0, 1), (10, 0), (10, 10, -0.5), (0, 10)], is_closed=True
    )
    paths.all_to_spline_edges(num=32)
    path = paths[0]

    edge = path.edges[0]
    assert edge.type == EdgeType.SPLINE
    assert edge.control_points[0].isclose((0, 0))
    assert edge.control_points[-1].isclose((10, 0))

    edge = path.edges[2]
    assert edge.type == EdgeType.SPLINE
    assert Vec3(10, 10).isclose(edge.control_points[0])
    assert Vec3(0, 10).isclose(edge.control_points[-1])


def add_edge_path(paths: BoundaryPaths):
    path = paths.add_edge_path()
    path.add_line((0, 0), (1, 0))
    path.add_line((1, 0), (1, 1))
    path.add_line((1, 1), (0, 1))
    path.add_line((0, 1), (0, 0))


def test_edge_to_polyline_paths():
    paths = BoundaryPaths()
    add_edge_path(paths)
    paths.edge_to_polyline_paths(distance=0.1)
    path = paths[0]
    assert path.type == BoundaryPathType.POLYLINE
    assert path.has_bulge() is False
    assert len(path.vertices) == 5


class TestValidatingBoundaryPaths:
    def test_spline_edge_has_valid_knot_count(self):
        paths = BoundaryPaths()
        edge_path = paths.add_edge_path()
        spline = edge_path.add_spline(
            control_points=[(0, 0), (1, 1), (2, 0), (3, 1)],
            knot_values=[1, 2, 3, 4, 5, 6, 7, 8],
        )
        assert spline.is_valid() is True, "expected degree + count + 1 values"

        spline.knot_values = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        assert spline.is_valid() is False, "too many knot value"
        spline.knot_values = [0, 0, 0, 0]
        assert spline.is_valid() is False, "too few knot value"

        assert paths.is_valid() is False, "invalid state should be propagated"

    def test_spline_edge_has_enough_control_points(self):
        paths = BoundaryPaths()
        edge_path = paths.add_edge_path()
        spline = edge_path.add_spline(
            control_points=[(0, 0), (1, 1), (2, 0)],
            knot_values=[1, 2, 3, 4, 5, 6, 7],  # correct degree + count + 1!
        )
        assert spline.is_valid() is False, "too few control points"

    def test_spline_edge_has_enough_fir_points(self):
        paths = BoundaryPaths()
        edge_path = paths.add_edge_path()
        spline = edge_path.add_spline(
            fit_points=[(0, 0), (1, 1), (2, 0)],
        )
        assert spline.is_valid() is True, "expected 2 or more points"
        spline.fit_points = [(0, 0)]
        assert spline.is_valid() is False, "too few fit points"


def test_real_start_point_of_arc_edge():
    paths = BoundaryPaths()
    edge_path = paths.add_edge_path()
    arc_edge = edge_path.add_arc(
        (1, 0), radius=1, start_angle=0, end_angle=180, ccw=True
    )
    assert arc_edge.real_start_point.isclose((2, 0))
    arc_edge = edge_path.add_arc(
        (1, 0), radius=1, start_angle=0, end_angle=180, ccw=False
    )
    assert arc_edge.real_start_point.isclose((0, 0))


def test_real_start_point_of_ellipse_edge():
    paths = BoundaryPaths()
    edge_path = paths.add_edge_path()
    arc_edge = edge_path.add_ellipse(
        (1, 0), major_axis=(1, 0), ratio=1, start_angle=0, end_angle=180, ccw=True
    )
    assert arc_edge.real_start_point.isclose((2, 0))
    arc_edge = edge_path.add_ellipse(
        (1, 0), major_axis=(1, 0), ratio=1, start_angle=0, end_angle=180, ccw=False
    )
    assert arc_edge.real_start_point.isclose((0, 0))


def test_real_end_point_of_arc_edge():
    paths = BoundaryPaths()
    edge_path = paths.add_edge_path()
    arc_edge = edge_path.add_arc(
        (1, 0), radius=1, start_angle=0, end_angle=180, ccw=True
    )
    assert arc_edge.real_end_point.isclose((0, 0))
    arc_edge = edge_path.add_arc(
        (1, 0), radius=1, start_angle=0, end_angle=180, ccw=False
    )
    assert arc_edge.real_end_point.isclose((2, 0))


def test_real_end_point_of_ellipse_edge():
    paths = BoundaryPaths()
    edge_path = paths.add_edge_path()
    arc_edge = edge_path.add_ellipse(
        (1, 0), major_axis=(1, 0), ratio=1, start_angle=0, end_angle=180, ccw=True
    )
    assert arc_edge.real_end_point.isclose((0, 0))
    arc_edge = edge_path.add_ellipse(
        (1, 0), major_axis=(1, 0), ratio=1, start_angle=0, end_angle=180, ccw=False
    )
    assert arc_edge.real_end_point.isclose((2, 0))


class TestCloseGapsOfEdgePaths:
    def test_gap_between_line_edges(self):
        paths = BoundaryPaths()
        edge_path = paths.add_edge_path()
        edge_path.add_line((0, 0), (1, 0))
        edge_path.add_line((1.1, 0), (1, 1))
        edge_path.add_line((1, 1), (0, 0))
        edge_path.close_gaps(0.01)
        assert len(edge_path.edges) == 4

    def test_gap_between_last_and_first_line(self):
        paths = BoundaryPaths()
        edge_path = paths.add_edge_path()
        edge_path.add_line((0, 0), (1, 0))
        edge_path.add_line((1, 0), (1, 1))
        edge_path.add_line((1, 1), (0, 0.1))
        edge_path.close_gaps(0.01)
        assert len(edge_path.edges) == 4

    def test_gap_between_arc_and_line_ccw(self):
        paths = BoundaryPaths()
        edge_path = paths.add_edge_path()
        edge_path.add_arc((1, 0), radius=1, start_angle=0, end_angle=180)
        edge_path.add_line((0, 0), (1.9, 0))
        edge_path.close_gaps(0.01)
        assert len(edge_path.edges) == 3

    def test_gap_between_arc_and_line_cw(self):
        paths = BoundaryPaths()
        edge_path = paths.add_edge_path()
        edge_path.add_arc((1, 0), radius=1, start_angle=0, end_angle=180, ccw=False)
        edge_path.add_line((1.9, 0), (0, 0))
        edge_path.close_gaps(0.01)
        assert len(edge_path.edges) == 3

    def test_gap_between_ellipse_and_line_ccw(self):
        paths = BoundaryPaths()
        edge_path = paths.add_edge_path()
        edge_path.add_ellipse(
            (1, 0), major_axis=(1, 0), ratio=1, start_angle=0, end_angle=180
        )
        edge_path.add_line((0, 0), (1.9, 0))
        edge_path.close_gaps(0.01)
        assert len(edge_path.edges) == 3

    def test_gap_between_ellipse_and_line_cw(self):
        paths = BoundaryPaths()
        edge_path = paths.add_edge_path()
        edge_path.add_ellipse(
            (1, 0), major_axis=(1, 0), ratio=1, start_angle=0, end_angle=180, ccw=False
        )
        edge_path.add_line((1.9, 0), (0, 0))
        edge_path.close_gaps(0.01)
        assert len(edge_path.edges) == 3


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License

import pytest
import math
import ezdxf
from ezdxf.upright import upright, _flip_deg_angle
from ezdxf import path
from ezdxf.entities import (
    Circle,
    Arc,
    DXFEntity,
    Text,
    Solid,
    Trace,
    Ellipse,
    LWPolyline,
    Hatch,
)
from ezdxf.math import Z_AXIS, Matrix44, OCS, OCSTransform, Vec3


@pytest.mark.parametrize(
    "angle",
    [
        0.0,
        30.0,
        60.0,
        90.0,
        180.0,
        270.0,
        -30.0,
        -60.0,
        -90.0,
        -180.0,
        -270.0,
    ],
)
def test_flip_deg_angle(angle):
    t = OCSTransform.from_ocs(
        OCS(-Z_AXIS),
        OCS(Z_AXIS),
        Matrix44(),
    )
    control_value = t.transform_deg_angle(angle)
    assert _flip_deg_angle(angle) == pytest.approx(control_value)


@pytest.fixture
def circle():
    return Circle.new(
        dxfattribs={"center": (3, 4), "radius": 2.0, "extrusion": (0, 0, -1)}
    )


def test_safety_checks(circle):
    # invalid entities should be ignored silently
    upright(None)  # ignore None values
    upright(DXFEntity())  # ignore invalid DXF entity types
    upright(Text())  # ignore unsupported DXF entity types
    circle.destroy()
    upright(circle)  # ignore destroyed entities
    assert True is True


def test_upright_circle_dxf_attributes(circle):
    upright(circle)
    assert circle.dxf.extrusion.isclose(Z_AXIS)
    assert circle.dxf.center.isclose((-3, 4))
    assert circle.dxf.radius == 2.0


def test_upright_circle_geometry(circle):
    circle.dxf.center = (0, 0)  # required for rotation!
    p0 = path.make_path(circle)
    upright(circle)
    # IMPORTANT: Circle has a different WCS representation as Path object
    # Rotated around the z-axis by 180 degrees AND reversed order, because
    # the start point is always at 0 degrees, determined by the OCS x-axis!
    p1 = path.make_path(circle).transform(Matrix44.z_rotate(math.pi))
    assert path.have_close_control_vertices(p0, p1.reversed())


@pytest.fixture
def arc():
    return Arc.new(
        dxfattribs={
            "center": (3, 4, 5),
            "radius": 2.0,
            "start_angle": 15,
            "end_angle": 75,
            "extrusion": (0, 0, -1),
        }
    )


def test_upright_arc_dxf_attributes(arc):
    upright(arc)
    assert arc.dxf.extrusion.isclose(Z_AXIS)
    assert arc.dxf.center.isclose((-3, 4, -5))
    assert arc.dxf.radius == 2.0
    assert arc.dxf.start_angle == pytest.approx(105.0)
    assert arc.dxf.end_angle == pytest.approx(165.0)


def test_upright_arc_geometry(arc):
    p0 = path.make_path(arc)
    upright(arc)
    # ARC angles are always in counter-clockwise orientation around the
    # extrusion vector, therefore a reversed path vertex order:
    p1 = path.make_path(arc).reversed()
    assert path.have_close_control_vertices(p0, p1)


@pytest.mark.parametrize("cls", [Solid, Trace])
def test_upright_quadrilaterals(cls):
    solid = cls.new(
        dxfattribs={
            "vtx0": (1, 1),
            "vtx1": (2, 1),
            "vtx2": (2, 2),
            "vtx3": (1, 2),
            "extrusion": (0, 0, -1),
        }
    )
    p0 = path.make_path(solid)
    assert len(p0) == 4

    upright(solid)
    assert solid.dxf.extrusion.isclose(Z_AXIS)
    p1 = path.make_path(solid)
    # same vertex order as source entity
    assert path.have_close_control_vertices(p0, p1)


def test_upright_ellipse():
    ellipse = Ellipse.new(
        dxfattribs={
            "center": (5, 5, 5),
            "major_axis": (5, 0, 0),
            "ratio": 0.5,
            "start_param": 0.5,
            "end_param": 1.5,
            "extrusion": (0, 0, -1),
        }
    )
    p0 = path.make_path(ellipse)
    assert p0.has_curves is True

    upright(ellipse)
    assert ellipse.dxf.extrusion.isclose(Z_AXIS)
    p1 = path.make_path(ellipse)
    # has reversed vertex order of source entity:
    assert path.have_close_control_vertices(p0, p1.reversed())


POLYLINE_POINTS = [
    # x, y, s, e, b
    (0, 0, 0, 0, 0),
    (2, 2, 1, 2, -1),
    (4, 0, 2, 1, 1),
    (6, 0, 0, 0, 0),
]


def lwpolyline():
    pline = LWPolyline.new(
        dxfattribs={
            "elevation": 4,
            "extrusion": (0, 0, -1),
        }
    )
    pline.set_points(POLYLINE_POINTS)
    return pline


def polyline2d():
    from ezdxf.layouts import VirtualLayout

    layout = VirtualLayout()
    return layout.add_polyline2d(
        POLYLINE_POINTS,
        format="xyseb",
        dxfattribs={
            "elevation": (0, 0, 4),
            "extrusion": (0, 0, -1),
        },
    )


@pytest.mark.parametrize("factory", [lwpolyline, polyline2d])
def test_upright_polyline(factory):
    polyline = factory()
    p0 = path.make_path(polyline)
    assert p0.has_curves is True

    upright(polyline)
    assert polyline.dxf.extrusion.isclose(Z_AXIS)
    p1 = path.make_path(polyline)
    # vertex order do not change:
    assert path.have_close_control_vertices(p0, p1)


def test_upright_hatch_with_polyline_path():
    hatch = Hatch.new(
        dxfattribs={
            "elevation": (0, 0, 4),
            "extrusion": (0, 0, -1),
        }
    )
    hatch.paths.add_polyline_path(
        [(x, y, b) for x, y, s, e, b in POLYLINE_POINTS]
    )
    p0 = path.make_path(hatch)
    assert p0.has_curves is True

    upright(hatch)
    assert hatch.dxf.extrusion.isclose(Z_AXIS)
    p1 = path.make_path(hatch)
    assert path.have_close_control_vertices(p0, p1)


def test_upright_hatch_with_edge_path(all_edge_types_hatch):
    hatch = all_edge_types_hatch
    hatch.dxf.elevation = Vec3(0, 0, 4)
    hatch.dxf.extrusion = Vec3(0, 0, -1)
    assert hatch.dxf.extrusion.isclose(-Z_AXIS)

    p0 = path.make_path(hatch)
    assert p0.has_curves is True

    upright(hatch)
    assert hatch.dxf.extrusion.isclose(Z_AXIS)
    p1 = path.make_path(hatch)
    assert path.have_close_control_vertices(p0, p1)


def test_upright_insert():
    doc = ezdxf.new()
    blk = doc.blocks.new("example")
    blk.add_arc(
        center=(5, 0, 2),
        radius=3,
        start_angle=30,
        end_angle=150,
    )
    blk.add_lwpolyline(POLYLINE_POINTS)
    msp = doc.modelspace()
    blk_ref = msp.add_blockref(
        name="example",
        insert=(0, 0, 4),
        dxfattribs={
            "extrusion": (0, 0, -1),
            "rotation": -37,
        },
    )
    blk_ref_copy = blk_ref.copy()
    upright(blk_ref_copy)
    msp.add_entity(blk_ref_copy)

    assert blk_ref_copy.dxf.extrusion.isclose(Z_AXIS)
    for e0, e1 in zip(
        blk_ref.virtual_entities(), blk_ref_copy.virtual_entities()
    ):
        assert e0.dxftype() == e1.dxftype(), "same DXF type expected"
        p0 = path.make_path(e0)
        assert len(p0) > 0, "source path cannot be empty"
        p1 = path.make_path(e1)
        assert len(p1) > 0, "upright path cannot be empty"
        assert path.have_close_control_vertices(
            p0, p1
        ), "expected same WCS representation"


if __name__ == "__main__":
    pytest.main([__file__])
